import epcpm._build

from ._version import get_versions
__version__ = get_versions()['version']
__sha__ = get_versions()['full-revisionid']
del get_versions

__version_tag__ = 'v{}'.format(__version__)
__build_tag__ = epcpm._build.job_id



import re

import attr
import lxml.etree
import xmldiff.diff
import xmldiff.main


@attr.s(frozen=True)
class ErrorLogEntry:
    line = attr.ib()
    message = attr.ib()
    path = attr.ib()
    original = attr.ib(cmp=False, hash=False, repr=False)

    @classmethod
    def from_lxml(cls, original):
        return cls(
            line=original.line,
            message=original.message,
            path=original.path,
            original=original,
        )


@attr.s
class ValidationResult:
    failed = attr.ib()
    notes = attr.ib(default='')


def validate_against_schema(subject, schema):
    subject_xml = lxml.etree.fromstring(subject.read_bytes())
    success = schema.validate(subject_xml)

    errors = tuple(
        ErrorLogEntry.from_lxml(error)
        for error in schema.error_log
    )

    if len(errors) == 0:
        return ValidationResult(
            failed=False,
            notes='No validation errors for subject',
        )

    return ValidationResult(
        failed=not success,
        notes='\n'.join(f'{error}' for error in errors),
    )


def validate_against_reference(subject, schema, reference):
    failed = False

    schema.validate(reference)
    reference_error_log = schema.error_log
    schema.validate(subject)
    subject_error_log = schema.error_log
    reference_errors, subject_errors = (
        {
            ErrorLogEntry.from_lxml(original=original)
            for original in log
        }
        for log in (reference_error_log, subject_error_log)
    )

    diff = compare_to_reference(subject=subject, reference=reference)

    notes = []

    if len(diff) > 0:
        failed = True

        notes.extend(diff)

    if reference_errors != subject_errors:
        failed = True

    if not failed:
        return ValidationResult(
            failed=False,
            notes='No different errors between subject and reference',
        )

    reference_only = reference_errors - subject_errors
    subject_only = subject_errors - reference_errors
    if len(reference_only) > 0:
        notes.append('')
        notes.append('reference only:')
        notes.extend(f'{error}' for error in sorted(reference_only))
    if len(subject_only) > 0:
        notes.append('')
        notes.append('  subject only:')
        notes.extend(f'{error}' for error in sorted(subject_only))

    return ValidationResult(
        failed=True,
        notes='\n'.join(notes)
    )


def get_change_attribute(change):
    return change.node.rpartition('/')[2].partition('[')[0]


def ignore_read_only_scale_factor_to_writable(change, reference):
    attrib_name = 'access'

    if isinstance(change, xmldiff.diff.UpdateAttrib):
        of_interest = (
            get_change_attribute(change) == 'point'
            and change.name == attrib_name
            and change.value == 'rw'
        )
        if of_interest:
            # TODO: https://github.com/Shoobx/xmldiff/issues/33
            #       This is presently hazardous because if a previous
            #       change would cause things to move around then
            #       this may grab the incorrect element from the
            #       reference.
            left, = reference.xpath(change.node)
            if left.attrib.get('type') == 'sunssf':
                return left.attrib[attrib_name] == 'r'

    return False


def ignore_var_we_do_not_like(change, reference):
    attrib_name = 'units'

    if isinstance(change, xmldiff.diff.UpdateAttrib):
        of_interest = (
            get_change_attribute(change) == 'point'
            and change.name == attrib_name
            and change.value == 'VAr'
        )
        if of_interest:
            left, = reference.xpath(change.node)
            return left.attrib[attrib_name].casefold() == 'var'

    return False


def context_is_vendor_specific(context):
    attributes = context.context_node.attrib

    return re.search(r'Vnd\d*$', attributes['id']) is not None


def vendor_specific_elements(tree):
    function_namespace = lxml.etree.FunctionNamespace(None)
    function_namespace['vendor_specific'] = context_is_vendor_specific

    return tree.xpath(
        "/sunSpecModels/model/block/point[vendor_specific()]/symbol",
    ) + tree.xpath(
        "/sunSpecModels/strings/point[vendor_specific()]/symbol",
    )


def remove_element(element):
    parent = element.getparent()
    parent.remove(element)
    if len(parent) == 0 and len(parent.text.strip()) == 0:
        parent.text = None


def remove_elements_by_name(tree, names):
    xpath = '//*[{}]'.format(
        ' or '.join(f'name() = "{name}"' for name in names),
    )
    points = tree.xpath(xpath)

    for point in points:
        remove_element(element=point)

    return tree


def compare_to_reference(subject, reference):
    trimmed_subject, trimmed_reference = (
        remove_elements_by_name(tree=tree, names=('description', 'notes'))
        for tree in (subject, reference)
    )

    for element in vendor_specific_elements(trimmed_subject):
        remove_element(element=element)

    diff = xmldiff.main.diff_trees(trimmed_reference, trimmed_subject)

    return tuple(
        f'{change}'
        for change in diff
        if not any(
            f(
                change=change,
                reference=trimmed_reference,
            )
            for f in (
                ignore_read_only_scale_factor_to_writable,
                ignore_var_we_do_not_like,
            )
        )
    )


@attr.s
class PairedPaths:
    pairs = attr.ib(factory=dict)
    only_left = attr.ib(factory=list)
    only_right = attr.ib(factory=list)

    @classmethod
    def from_directories(cls, left_path, right_path, file_glob):
        left_paths = left_path.glob(file_glob)
        right_paths = right_path.glob(file_glob)

        left_relative_paths = {
            path.relative_to(left_path)
            for path in left_paths
        }

        right_relative_paths = {
            path.relative_to(right_path)
            for path in right_paths
        }

        only_left = {
            left_path/path
            for path in left_relative_paths - right_relative_paths
        }
        only_right = {
            right_path/path
            for path in right_relative_paths - left_relative_paths
        }
        common = left_relative_paths.intersection(right_relative_paths)

        paired_paths = {
            left_path/path: right_path/path
            for path in common
        }

        return cls(
            pairs=paired_paths,
            only_left=only_left,
            only_right=only_right,
        )



import collections
import decimal
import functools
import itertools
import json
import pathlib
import re
import uuid

import attr
import canmatrix.formats

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.canmodel
import epcpm.sunspecmodel


imported_variants = []

def humanize_name(name):
    return name
#     name = name.replace('_', ' - ')
#     return epyqlib.utils.general.underscored_camel_to_title_spaced(name)


def load_can_path(can_path, hierarchy_path):
    with open(can_path, 'rb') as c, open(hierarchy_path) as h:
        return load_can_file(
            can_file=c,
            file_type=str(pathlib.Path(can_path).suffix[1:]),
            parameter_hierarchy_file=h,
        )


def get_other_name(hierarchy):
    other_names = [
        child['name']
        for child in hierarchy['children']
        if isinstance(child, dict) and child.get('unreferenced', False)
    ]

    if not other_names:
        other_name = '9. Other'
    else:
        other_name, = other_names

    return other_name


def load_can_file(
        can_file,
        file_type,
        parameter_hierarchy_file,
        add_tables=False,
):
    matrix, = canmatrix.formats.load(
        can_file,
        file_type,
        calc_min_for_none=False,
        calc_max_for_none=False,
        float_factory=decimal.Decimal,
    ).values()

    parameters_root = epyqlib.pm.parametermodel.Root()
    can_root = epcpm.canmodel.Root()

    enumerations = epyqlib.pm.parametermodel.Enumerations(name='Enumerations')
    parameters_root.append_child(enumerations)

    parameters = epyqlib.pm.parametermodel.Group(name='Parameters')
    parameters_root.append_child(parameters)

    parameter_hierarchy = json.load(
        parameter_hierarchy_file,
        object_pairs_hook=collections.OrderedDict,
    )

    other_name = get_other_name(parameter_hierarchy)
    other_parameters = epyqlib.pm.parametermodel.Group(name=other_name)
    parameters.append_child(other_parameters)

    ccp = epyqlib.pm.parametermodel.Group(name='CCP')
    parameters_root.append_child(ccp)

    process_to_inverter = epyqlib.pm.parametermodel.Group(
        name='Process To Inverter',
    )
    parameters_root.append_child(process_to_inverter)

    other = epyqlib.pm.parametermodel.Group(name='9. Other')
    parameters_root.append_child(other)

    read_nvs = epyqlib.pm.parametermodel.Group(name='ReadNV')
    other.append_child(read_nvs)

    nv_commands = epyqlib.pm.parametermodel.Group(name='NVCommands')
    other.append_child(nv_commands)

    ccp_counters = epyqlib.pm.parametermodel.Group(name='CCP Counters')
    other.append_child(ccp_counters)

    def traverse_hierarchy(children, parent, group_from_path):
        for child in children:
            if isinstance(child, dict):
                subchildren = child.get('children')
                if subchildren is None:
                    continue

                existing = {
                    group.name: group
                    for group in parent.children
                }

                group = existing.get(child['name'])
                if group is None:
                    group = epyqlib.pm.parametermodel.Group(
                        name=child['name'],
                    )
                    parent.append_child(group)

                traverse_hierarchy(
                    children=subchildren,
                    parent=group,
                    group_from_path=group_from_path,
                )
            else:
                group_from_path[('ParameterQuery',) + tuple(child)] = parent
                group_from_path[('ParameterResponse',) + tuple(child)] = parent

    group_from_path = {
        'other_parameters': other_parameters,
        'read_nvs': read_nvs,
        'nv_commands': nv_commands,
        'ccp': ccp,
        'ccp_counters': ccp_counters,
        'process_to_inverter': process_to_inverter,
        'other': other,
    }
    traverse_hierarchy(
        children=parameter_hierarchy['children'],
        parent=parameters,
        group_from_path=group_from_path,
    )

    enumeration_name_to_uuid = {}
    access_levels = None
    variants = None
    for name, values in sorted(matrix.value_tables.items()):
        if name == 'AccessLevel':
            enumeration = epyqlib.pm.parametermodel.AccessLevels(name=name)
            access_levels = enumeration
            enumerator_type = epyqlib.pm.parametermodel.AccessLevel
        else:
            enumeration = epyqlib.pm.parametermodel.Enumeration(name=name)
            enumerator_type = epyqlib.pm.parametermodel.Enumerator

        enumerations.append_child(enumeration)
        enumeration_name_to_uuid[name] = enumeration.uuid

        for value, gname in values.items():
            enumerator = enumerator_type(name=gname, value=value)
            enumeration.append_child(enumerator)

        if name == 'CmmControlsVariant':
            variants = [
                variant 
                for variant in enumeration.children 
                if variant.name != 'None'
            ]
            global imported_variants
            imported_variants = variants

    parameter_from_path = {}

    for frame in matrix.frames:
        if len(frame.mux_names) == 0:
            message = build_message(
                frame=frame,
                parameter_group=parameters_root,
                enumeration_name_to_uuid=enumeration_name_to_uuid,
                access_levels=access_levels,
                variants=variants,
            )
        else:
            message = build_multiplexed_message(
                enumeration_name_to_uuid=enumeration_name_to_uuid,
                frame=frame,
                group_from_path=group_from_path,
                parameter_from_path=parameter_from_path,
                access_levels=access_levels,
                variants=variants,
            )

            def interesting_signal_attributes(signal):
                return (
                     signal.name,
                     signal.bits,
                     signal.signed,
                     signal.factor,
                     signal.start_bit,
                 )

            multiplexers = [
                mux
                for mux in message.children
                if isinstance(mux, epcpm.canmodel.Multiplexer)
            ]

            common_signals_dict = functools.reduce(
                epyqlib.utils.general.intersect_dicts,
                 (
                     {
                         interesting_signal_attributes(signal): signal
                         for signal in mux.children
                     }
                     for mux in multiplexers
                 ),
            )

            common_signals = list(common_signals_dict.values())
            common_signals_check = list(common_signals_dict.keys())

            # TODO: maybe coming back...  maybe
            # for mux in multiplexers:
            #     for child in list(mux.children):
            #         if interesting_signal_attributes(child) in common_signals_check:
            #             mux.remove_child(child=child)

            for signal in reversed(common_signals):
                message.insert_child(1, attr.evolve(signal, uuid=attr.NOTHING))

        can_root.append_child(message)

    def reorder_children(node, names):
        children_by_name = {
            child.name: child
            for child in node.children
        }

        for name in names:
            child = children_by_name[name]
            node.remove_child(child=child)
            node.append_child(child)

    def traverse_hierarchy_to_reorder(children, parent):
        for child in children:
            if isinstance(child, dict):
                child_objects = parent.children_by_attribute(
                    name='name',
                    value=child['name'],
                )
                if len(child_objects) == 1:
                    child_object, = child_objects
                else:
                    # TODO: just get rid of this debugging if and adjust above
                    raise Exception()

                subchildren = child.get('children')
                if subchildren is None:
                    continue

                subchild_names = []
                for subchild in subchildren:
                    if isinstance(subchild, dict):
                        name = subchild['name']
                    else:
                        # Query and response point the same place
                        signal_path = ('ParameterQuery',) + tuple(subchild)
                        name = parameter_from_path[signal_path].name

                    subchild_names.append(name)

                reorder_children(node=child_object, names=subchild_names)

                traverse_hierarchy_to_reorder(
                    children=subchildren,
                    parent=child_object,
                )

    traverse_hierarchy_to_reorder(
        children=parameter_hierarchy['children'],
        parent=parameters,
    )

    sunspec_root = epcpm.sunspecmodel.Root()

    def stripped(parameter):
        prefix = '{} : '.format(parameter.original_multiplexer_name)
        if parameter.name.startswith(prefix):
            return parameter.name[len(prefix):]

        return parameter.name

    def strip_frame_name(node, payload):
        if isinstance(node, epyqlib.pm.parametermodel.Parameter):
            other_names = {
                stripped(other)
                for other in node.tree_parent.children
                if (
                    isinstance(other, epyqlib.pm.parametermodel.Parameter)
                    and other is not node
                )
            }
            stripped_name = stripped(node)
            if stripped_name not in other_names:
                node.name = stripped_name

    parameters_root.traverse(
        call_this=strip_frame_name,
        payload=None,
        internal_nodes=True,
    )

    return parameters_root, can_root, sunspec_root


def build_message(
        frame,
        parameter_group,
        enumeration_name_to_uuid,
        access_levels,
        variants,
):
    extras = {}

    if frame.cycle_time is not None:
        extras['cycle_time'] = frame.cycle_time

    message = message_from_matrix(
        frame=frame,
        factory=epcpm.canmodel.Message,
        length=frame.size,
        **extras,
    )
    group = epyqlib.pm.parametermodel.Group(
        name=humanize_name(frame.name),
    )
    parameter_group.append_child(group)
    
    variant_cfgs = strip_variant_parameter_tag(string='', variants=variants)[1]

    for matrix_signal in frame.signals:
        parameter = parameter_from_signal(
            frame=frame,
            frame_access_level=access_levels.default(),
            matrix_signal=matrix_signal,
            enumeration_name_to_uuid=enumeration_name_to_uuid,
            access_levels=access_levels,
            variants=variants,
            frame_variants=variant_cfgs
        )
        group.append_child(parameter)

        signal = signal_from_matrix(
            matrix_signal=matrix_signal,
            factory=epcpm.canmodel.Signal,
            enumeration_name_to_uuid=enumeration_name_to_uuid,
            parameter_uuid=parameter.uuid,
        )
        message.append_child(signal)

    return message


def message_from_matrix(frame, factory, **extras):
    extras.setdefault('name', humanize_name(frame.name))
    extras.setdefault('identifier', frame.arbitration_id.id)
    extras.setdefault('extended', frame.arbitration_id.extended)
    extras.setdefault('comment', frame.comment)
    extras.setdefault('sendable', frame.attributes['Sendable'] == 'True')
    extras.setdefault('receivable', frame.attributes['Receivable'] == 'True')

    return factory(
        **extras
    )


def signal_from_matrix(
        matrix_signal,
        factory,
        enumeration_name_to_uuid,
        **extras,
):
    extras.setdefault('name', humanize_name(matrix_signal.name))
    extras.setdefault('bits', matrix_signal.size)
    extras.setdefault('factor', matrix_signal.factor)
    extras.setdefault('signed', matrix_signal.is_signed)
    extras.setdefault('start_bit', matrix_signal.get_startbit())

    extras.setdefault(
        'enumeration_uuid',
        enumeration_name_to_uuid.get(matrix_signal.enumeration),
    )

    return factory(
        **extras
    )


def build_multiplexed_message(
        enumeration_name_to_uuid,
        frame,
        group_from_path,
        parameter_from_path,
        access_levels,
        variants,
):
    message = message_from_matrix(
        frame=frame,
        factory=epcpm.canmodel.MultiplexedMessage,
    )
    matrix_mux_signal, = (
        s
        for s in frame.signals
        if s.multiplex == 'Multiplexor'
    )
    mux_signal = signal_from_matrix(
        factory=epcpm.canmodel.Signal,
        matrix_signal=matrix_mux_signal,
        enumeration_name_to_uuid=enumeration_name_to_uuid,
        signed=False,
    )
    message.append_child(mux_signal)
    for value, mux_name in sorted(frame.mux_names.items()):
        extras = {}

        mux_comment = matrix_mux_signal.comments.get(value)
        if mux_comment is not None:
            mux_comment, access_level = strip_access_level(
                string=mux_comment,
                access_levels=access_levels,
            )
            
            mux_comment, variant_cfgs = strip_variant_parameter_tag(
                string=mux_comment,
                variants=variants,
            )

            if len(mux_comment) > 0:
                extras['comment'] = mux_comment

        if frame.cycle_time != 0:
            extras['cycle_time'] = frame.cycle_time

        multiplexer = epcpm.canmodel.Multiplexer(
            name=humanize_name(mux_name),
            identifier=value,
            length=frame.size,
            **extras,
        )
        message.append_child(multiplexer)

        for matrix_signal in frame.signals:
            if matrix_signal.multiplex != value:
                continue

            parameter = parameter_from_signal(
                frame=frame,
                frame_access_level=access_level,
                matrix_signal=matrix_signal,
                mux_name=mux_name,
                enumeration_name_to_uuid=enumeration_name_to_uuid,
                access_levels=access_levels,
                variants=variants,
                frame_variants=variant_cfgs,
            )

            group = group_from_path.get(
                (frame.name, mux_name, matrix_signal.name),
            )

            if group is None:
                if frame.name.startswith('Parameter'):
                    if matrix_signal.enumeration in {'ReadNV', 'Meta'}:
                        group = 'read_nvs'
                    elif matrix_signal.name.startswith('SaveToEE'):
                        group = 'nv_commands'
                    else:
                        group = 'other_parameters'
                elif frame.name.startswith('CCP'):
                    if matrix_signal.name == 'CommandCounter':
                        group = 'ccp_counters'
                    else:
                        group = 'ccp'
                elif frame.name.startswith('Process'):
                    group = 'process_to_inverter'
                else:
                    group = 'other'

                group = group_from_path[group]

            same_name_parameters = [
                node
                for node in group.children
                if node.name == parameter.name
            ]

            if len(same_name_parameters) == 0:
                group.append_child(parameter)
            else:
                parameter, = same_name_parameters

            signal_path = (frame.name, mux_name, matrix_signal.name)
            parameter_from_path[signal_path] = parameter

            signal = signal_from_matrix(
                matrix_signal=matrix_signal,
                factory=epcpm.canmodel.Signal,
                enumeration_name_to_uuid=enumeration_name_to_uuid,
                parameter_uuid=parameter.uuid,
            )

            multiplexer.append_child(signal)

    return message


# TODO: backmatching
def enumeration_definition(name, value_names):
    enumeration = epyqlib.pm.parametermodel.Enumeration(
        name=name,
    )
    for value, value_name in enumerate(value_names):
        enumerator = epyqlib.pm.parametermodel.Enumerator(
            name=value_name,
            value=value,
        )
        enumeration.append_child(enumerator)

    return enumeration


def array_definition(name, length, parameter):
    array = epyqlib.pm.parametermodel.Array(
        name=name,
    )
    array.append_child(parameter)
    array.length = length

    name_length = len('{}'.format(length))

    for i, node in enumerate(array.children, start=1):
        node.name = '_{:0{}}'.format(i, name_length)

    return array


def group_definition(name, parameters):
    group = epyqlib.pm.parametermodel.Group(
        name=name,
    )
    for parameter in parameters:
        group.append_child(parameter)

    return group


def table_definition(
        parent,
        name,
        enumerations,
        can_getter,
        can_setter,
        active_curve_getter,
        active_curve_setter,
        arrays,
        groups=(),
):
    table = epyqlib.pm.parametermodel.Table(
        name=name,
        can_getter=can_getter,
        can_setter=can_setter,
        active_curve_getter=active_curve_getter,
        active_curve_setter=active_curve_setter,
    )
    parent.append_child(table)

    for enumeration in enumerations:
        reference = epyqlib.pm.parametermodel.TableEnumerationReference(
            name=enumeration.name,
            enumeration_uuid=enumeration.uuid,
        )

        table.append_child(reference)

    for group in groups:
        table.append_child(group)

    for array in arrays:
        table.append_child(array)

    return table


@attr.s(frozen=True, slots=True)
class CanTableDefinition:
    signals = attr.ib()
    range = attr.ib()


def go_add_tables(parameters_root, can_root):
    def is_a_table(node):
        comment = getattr(node, 'comment', None)
        if comment is None:
            return False

        return '<table>' in comment

    can_table_stuff = can_root.nodes_by_filter(filter=is_a_table)
    for node in can_table_stuff:
        node.tree_parent.remove_child(child=node)

    line_monitoring = parameters_root.descendent(
        'Parameters',
        '1. AC',
        '10. Line Monitoring',
    )
    enumerations_group = parameters_root.descendent('Enumerations')
    existing_tables = line_monitoring.descendent('Tables')
    line_monitoring.remove_child(child=existing_tables)

    low_high = enumeration_definition(
        name='LowHigh',
        value_names=(
            'Low',
            'High',
        )
    )

    ridethrough_trip = enumeration_definition(
        name='RideThroughTrip',
        value_names=(
            'RideThrough',
            'Trip',
        )
    )

    curves = enumeration_definition(
        name='Curves',
        value_names=(
            '0',
            '1',
            '2',
            '3',
        )
    )

    volt_var_y_scale = enumeration_definition(
        name='VoltVarYScale',
        value_names=(
            "rated reactive power with vars precedence",
            "rated reactive power with watts precedence",
            "available reactive power with watts precedence",
        ),
    )

    enumerations = (
        low_high,
        ridethrough_trip,
        curves,
        volt_var_y_scale
    )

    for enumeration in enumerations:
        enumerations_group.append_child(enumeration)

    curve_points = 10

    tables_group = epyqlib.pm.parametermodel.Group(
        name='Tables',
    )
    line_monitoring.append_child(tables_group)

    frequency_table = table_definition(
        parent=tables_group,
        name='Frequency',
        enumerations=(
            low_high,
            ridethrough_trip,
            curves,
        ),
        can_getter='{interface_signal} = lineMonitorParams->fMonitorTables[{curve_type}].curves[{curve_index}].tbl[{point_index}].{axis};',
        can_setter='gridMonitor_setFrequencyZoneCurve{upper_axis}Point({curve_type}, {curve_index}, {point_index}, {interface_signal});',
        active_curve_getter='{interface_signal} = lineMonitorParams->fMonitorTables[{curve_type}].activeCurve;',
        active_curve_setter='gridMonitor_setFrequencyZoneActiveCurve({curve_type}, {interface_signal});',
        arrays=(
            array_definition(
                name='seconds',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='seconds',
                    decimal_places=2,
                ),
            ),
            array_definition(
                name='hertz',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='Hz',
                    decimal_places=2,
                ),
            ),
        ),
    )

    voltage_table = table_definition(
        parent=tables_group,
        name='Voltage',
        enumerations=(
            low_high,
            ridethrough_trip,
            curves,
        ),
        can_getter='{interface_signal} = lineMonitorParams->vMonitorTables[{curve_type}].curves[{curve_index}].tbl[{point_index}].{axis};',
        can_setter='gridMonitor_setVoltageZoneCurve{upper_axis}Point({curve_type}, {curve_index}, {point_index}, {interface_signal});',
        active_curve_getter='{interface_signal} = lineMonitorParams->vMonitorTables[{curve_type}].activeCurve;',
        active_curve_setter='gridMonitor_setVoltageZoneActiveCurve({curve_type}, {interface_signal});',
        arrays=(
            array_definition(
                name='seconds',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='seconds',
                    decimal_places=2,
                ),
            ),
            array_definition(
                name='percent',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='%',
                    decimal_places=1,
                ),
            ),
        ),
    )

    settings = group_definition(
        name='Settings',
        parameters=(
            epyqlib.pm.parametermodel.Parameter(
                name='YScale',
                maximum=3,
                can_getter='{interface_signal} = lineMonitorParams->voltVar.yScale[{curve_index}];',
                can_setter='lineMonitorParams->voltVar.yScale[{curve_index}] = (VoltVar_YScale) {interface_signal};',
            ),
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateIncrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->voltVar.droopTable.curves[{curve_index}].rampRateInc;',
                can_setter='gridMonitor_setVoltVarCurveRampInc({curve_index}, {interface_signal});',
            ),
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateDecrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->voltVar.droopTable.curves[{curve_index}].rampRateDec;',
                can_setter='gridMonitor_setVoltVarCurveRampDec({curve_index}, {interface_signal});',
            ),
        ),
    )

    volt_var_table = table_definition(
        parent=tables_group,
        name='VoltVar',
        enumerations=(curves,),
        can_getter='{interface_signal} = lineMonitorParams->voltVar.droopTable.curves[{curve_index}].tbl[{point_index}].{axis};',
        can_setter='gridMonitor_setVoltVarCurve{upper_axis}Point({curve_index}, {point_index}, {interface_signal});',
        active_curve_getter='{interface_signal} = lineMonitorParams->voltVar.droopTable.activeCurve;',
        active_curve_setter='gridMonitor_setVoltVarActiveCurve({interface_signal});',
        arrays=(
            settings,
            array_definition(
                name='percent_nominal_volts',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='% nominal V',
                    decimal_places=1,
                ),
            ),
            array_definition(
                name='percent_nominal_var',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='% nominal VAr',
                    decimal_places=1,
                ),
            ),
        ),
    )

    settings = group_definition(
        name='Settings',
        parameters=(
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateIncrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->hzWatts.curves[{curve_index}].rampRateInc;',
                can_setter='gridMonitor_setHzWattsCurveRampInc({curve_index}, {interface_signal});',
            ),
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateDecrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->hzWatts.curves[{curve_index}].rampRateDec;',
                can_setter='gridMonitor_setHzWattsCurveRampDec({curve_index}, {interface_signal});',
            ),
        ),
    )

    hertz_watts_table = table_definition(
        parent=tables_group,
        name='HzWatts',
        enumerations=(curves,),
        can_getter='{interface_signal} = lineMonitorParams->hzWatts.curves[{curve_index}].tbl[{point_index}].{axis};',
        can_setter='gridMonitor_setHzWattsCurve{upper_axis}Point({curve_index}, {point_index}, {interface_signal});',
        active_curve_getter='{interface_signal} = lineMonitorParams->hzWatts.activeCurve;',
        active_curve_setter='gridMonitor_setHzWattsActiveCurve({interface_signal});',
        groups=(settings,),
        arrays=(
            array_definition(
                name='hertz',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='Hz',
                    decimal_places=2,
                ),
            ),
            array_definition(
                name='percent_nominal_pwr',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='% nominal Power',
                    decimal_places=1,
                ),
            ),
        ),
    )

    settings = group_definition(
        name='Settings',
        parameters=(
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateIncrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->voltWatts.curves[{curve_index}].rampRateInc;',
                can_setter='gridMonitor_setVoltWattsCurveRampInc({curve_index}, {interface_signal});',
            ),
            epyqlib.pm.parametermodel.Parameter(
                name='RampRateDecrement',
                units='%/minute',
                can_getter='{interface_signal} = lineMonitorParams->voltWatts.curves[{curve_index}].rampRateDec;',
                can_setter='gridMonitor_setVoltWattsCurveRampDec({curve_index}, {interface_signal});',
            ),
        ),
    )

    volt_watts_table = table_definition(
        parent=tables_group,
        name='VoltWatts',
        enumerations=(curves,),
        can_getter='{interface_signal} = lineMonitorParams->voltWatts.curves[{curve_index}].tbl[{point_index}].{axis};',
        can_setter='gridMonitor_setVoltWattsCurve{upper_axis}Point({curve_index}, {point_index}, {interface_signal});',
        active_curve_getter='{interface_signal} = lineMonitorParams->voltWatts.activeCurve;',
        active_curve_setter='gridMonitor_setVoltWattsActiveCurve({interface_signal});',
        groups=(settings,),
        arrays=(
            array_definition(
                name='percent_nominal_volts',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='% nominal V',
                    decimal_places=1,
                ),
            ),
            array_definition(
                name='percent_nominal_pwr',
                length=curve_points,
                parameter=epyqlib.pm.parametermodel.Parameter(
                    units='% nominal Power',
                    decimal_places=1,
                ),
            ),
        ),
    )

    tables = {
        frequency_table: CanTableDefinition(
            range=range(0x195, 0x215),
            signals={
                'seconds': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.01',
                },
                'hertz': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.01',
                },
            },
        ),
        voltage_table: CanTableDefinition(
            range=range(0x215, 0x295),
            signals={
                'seconds': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.01',
                },
                'percent': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.1',
                },
            },
        ),
        volt_var_table: CanTableDefinition(
            range=range(0x295, 0x2b9),
            signals={
                'YScale': {
                    'bits': 2,
                    'signed': False,
                    'enumeration_uuid': volt_var_y_scale.uuid,
                },
                'RampRateIncrement': {
                    'bits': 16,
                    'signed': False,
                },
                'RampRateDecrement': {
                    'bits': 16,
                    'signed': False,
                },
                'percent_nominal_volts': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.1',
                },
                'percent_nominal_var': {
                    'bits': 16,
                    'signed': True,
                    'factor': '0.1',
                },
            },
        ),
        hertz_watts_table: CanTableDefinition(
            range=range(0x2b9, 0x2dd),
            signals={
                'RampRateIncrement': {
                    'bits': 16,
                    'signed': False,
                    'factor': '1',
                },
                'RampRateDecrement': {
                    'bits': 16,
                    'signed': False,
                    'factor': '1',
                },
                'hertz': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.01',
                },
                'percent_nominal_pwr': {
                    'bits': 16,
                    'signed': True,
                    'factor': '0.1',
                },
            },
        ),
        volt_watts_table: CanTableDefinition(
            range=range(0x2dd, 0x300),
            signals={
                'RampRateIncrement': {
                    'bits': 16,
                    'signed': False,
                    'factor': '1',
                },
                'RampRateDecrement': {
                    'bits': 16,
                    'signed': False,
                    'factor': '1',
                },
                'percent_nominal_volts': {
                    'bits': 16,
                    'signed': False,
                    'factor': '0.1',
                },
                'percent_nominal_pwr': {
                    'bits': 16,
                    'signed': True,
                    'factor': '0.1',
                },
            },
        ),
    }

    for table, can_table_definition in tables.items():
        table.update()

        for name in ('ParameterQuery', 'ParameterResponse'):
            can_group = can_root.descendent(name)

            can_table = can_group.child_from(table)
            can_table.name = table.name
            can_table.multiplexer_range_first = can_table_definition.range[0]
            can_table.multiplexer_range_second = can_table_definition.range[-1]
            can_group.append_child(can_table)
            can_table.update(warn=False)

            for signal, values in can_table_definition.signals.items():
                signal = can_table.descendent(signal)
                for name, value in values.items():
                    setattr(signal, name, value)

            can_table.update()

    active_curve_names = (
        'FrequencyLowRideThrough',
        'FrequencyHighRideThrough',
        'FrequencyLowTrip',
        'FrequencyHighTrip',
        'VoltageLowRideThrough',
        'VoltageHighRideThrough',
        'VoltageLowTrip',
        'VoltageHighTrip',
        'VoltVar',
        'HzWatts',
        'VoltWatts',
    )

    generated_active_curve_names = tuple(
        table.name + ''.join(enumerator.name for enumerator in combination)
        for table in tables
        for combination in table.curve_group_combinations
    )

    # TODO: backmatching
    if set(active_curve_names) != set(generated_active_curve_names):
        raise Exception(
            '{}, {}'.format(active_curve_names, generated_active_curve_names),
        )

    start_bit = 16
    bits = 4

    for name in ('ParameterQuery', 'ParameterResponse'):
        can_group = can_root.descendent(name)

        multiplexer = epcpm.canmodel.Multiplexer(
            name='ActiveCurves',
            identifier=can_table.children[-1].identifier + 1,
            comment='<table>'
        )
        can_group.append_child(multiplexer)

    for name in active_curve_names:
        parameter = epyqlib.pm.parametermodel.Parameter(
            name=name,
            maximum=3,
        )
        tables_group.append_child(parameter)

        for name in ('ParameterQuery', 'ParameterResponse'):
            active_curves_multiplexer = can_root.descendent(name, 'ActiveCurves')

            signal = active_curves_multiplexer.child_from(parameter)
            signal.bits = bits
            signal.start_bit = start_bit

            active_curves_multiplexer.append_child(signal)

        start_bit += bits


def strip_tag(string, tag):
    present = tag in string

    if present:
        string = string.replace(tag, '').strip()

    return string, present


def strip_access_level(string, access_levels):
    factory_tag = '<factory>'

    string, present = strip_tag(string, factory_tag)

    access_level = access_levels.default()

    if present:
        access_level = access_levels.by_name('factory')

    return string, access_level


def strip_variant_parameter_tag(string, variants):
    selected_variants = []
    
    for variant in variants:
        string, present = strip_tag(string, f'<{variant.name}>')
        if present:
            selected_variants.append(variant)
    
    if len(selected_variants) == 0:
        selected_variants = variants
        
    return string, selected_variants


@attr.s
class NvMeta:
    format = attr.ib()
    factor = attr.ib(default=None)
    cast = attr.ib(default=False)


nv_pattern = re.compile('<nv:(.*?)>')


def strip_nv(string):
    tags = nv_pattern.search(string)

    if tags is None:
        return string, None

    flags, *tags, format = tags[1].split(':')

    extras = {}

    flags = set(flags)
    if 'c' in flags:
        flags.remove('c')
        extras['cast'] = True

    if len(flags) > 0:
        raise Exception('Unknown flags found {}'.format(''.join(sorted(flags))))

    for tag in tags:
        if tag.startswith('f'):
            extras['factor'] = tag[1:]
        else:
            raise Exception('Unknown tag found {}'.format(repr(tag)))

    return (
        nv_pattern.sub('', string),
        NvMeta(
            format=format,
            **extras,
        ),
    )


def parameter_from_signal(
        frame,
        frame_access_level,
        matrix_signal,
        enumeration_name_to_uuid,
        access_levels,
        variants,
        frame_variants,
        mux_name=None,
):
    extras = {}

    attributes = matrix_signal.attributes

    signal_name = attributes.get('LongName')
    if signal_name is None:
        if mux_name is not None:
            signal_name = '{} : {}'.format(
                humanize_name(mux_name),
                humanize_name(matrix_signal.name),
            )
        else:
            signal_name = humanize_name(matrix_signal.name)

    hexadecimal = matrix_signal.attributes.get('HexadecimalOutput')
    if hexadecimal is not None:
        extras['display_hexadecimal'] = hexadecimal == 'True'

    if matrix_signal.min is not None:
        extras['minimum'] = matrix_signal.min

    if matrix_signal.max is not None:
        extras['maximum'] = matrix_signal.max

    access_level = access_levels.default()

    if matrix_signal.comment is not None:
        comment = matrix_signal.comment
        comment, signal_access_level = strip_access_level(
            string=comment,
            access_levels=access_levels,
        )
        
        comment, variant_cfgs = strip_variant_parameter_tag(
            string=comment,
            variants=variants,
        )

        #only variants in both lists:
        vis_list = list(set(frame_variants).intersection(variant_cfgs))
        extras['visibility'] = vis_list
        
        comment, nv_meta = strip_nv(string=comment)

        folded = matrix_signal.name.casefold()

        if folded.startswith('readparam') or folded == 'meta':
            access_level = access_levels.default()
        else:
            access_level = max(
                (
                    access_level,
                    signal_access_level,
                    frame_access_level,
                ),
                key=lambda x: x.value,
            )

            if nv_meta is not None:
                extras['nv_format'] = nv_meta.format
                extras['nv_factor'] = nv_meta.factor
                extras['nv_cast'] = nv_meta.cast

        if len(comment) > 0:
            extras['comment'] = comment

    if access_level is not None:
        extras['access_level_uuid'] = access_level.uuid

    if matrix_signal.unit is not None:
        if len(matrix_signal.unit) > 0:
            extras['units'] = matrix_signal.unit

    default = matrix_signal.initial_value
    if default is not None:
        if matrix_signal.factor is not None:
            # TODO: it seems this shouldn't be needed...  0754397432978432
            default *= decimal.Decimal(matrix_signal.factor)
        extras['default'] = default

    decimal_places = attributes.get('DisplayDecimalPlaces')
    if decimal_places is not None:
        extras['decimal_places'] = decimal_places

    if mux_name is not None:
        extras['original_multiplexer_name'] = mux_name

    return epyqlib.pm.parametermodel.Parameter(
        name=signal_name,
        original_frame_name=frame.name,
        original_signal_name=matrix_signal.name,
        **extras,
    )



import logging
import os.path
import sys

import click
from PyQt5 import QtCore, QtGui, QtWidgets

import epyqlib.utils.qt
import epcpm.mainwindow

import epcpm.cli.utils


# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


def main(project, verbosity, logger):
    app = QtWidgets.QApplication(sys.argv)

    epyqlib.utils.qt.exception_message_box_register_versions(
        version_tag=epcpm.__version_tag__,
        build_tag=epcpm.__build_tag__,
    )

    sys.excepthook = epyqlib.utils.qt.exception_message_box

    os_signal_timer = epyqlib.utils.qt.setup_sigint()

    QtCore.qInstallMessageHandler(epyqlib.utils.qt.message_handler)

    app.setStyleSheet('QMessageBox {{ messagebox-text-interaction-flags: {}; }}'
                      .format(QtCore.Qt.TextBrowserInteraction))

    app.setOrganizationName('EPC Power Corp.')
    app.setApplicationName('EPC Parameter Management')

    if verbosity >= 1:
        logger.setLevel(logging.DEBUG)

        if verbosity >= 2:
            # twisted.internet.defer.setDebugging(True)
            pass

            if verbosity >= 3:
                logging.getLogger().setLevel(logging.DEBUG)

    window = epcpm.mainwindow.Window(
        title='EPC Parameter Manager',
        icon_path='icon.ico',
    )

    epyqlib.utils.qt.exception_message_box_register_parent(parent=window.ui)

    if project is not None:
        filename = os.path.abspath(project)
        window.open_project(filename=filename)

    window.main_window.show()

    return app.exec()


@click.command()
@epcpm.cli.utils.project_option()
@click.option(
    '-v',
    '--verbose',
    'verbosity',
    count=True,
    help='Increase verbosity of output (up to three times)',
)
def _entry_point(project, verbosity):
    """Parameter Manager GUI"""

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    stream_handler = logging.StreamHandler()
    file_handler = logging.FileHandler('epcpm.log')

    for handler in (stream_handler, file_handler):
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s')

    sys.excepthook = epyqlib.utils.general.exception_logger

    return main(project=project, verbosity=verbosity, logger=logger)


# for PyInstaller
if __name__ == '__main__':
    sys.exit(_entry_point())



import epcpm.sunspectomanualc


import functools

import pycparser.c_parser


def parse(c):
    parser = pycparser.c_parser.CParser()
    ast = parser.parse(c)
    return ast

c = """
void getSUNSPEC_MODEL1_DA (void) {
    sunspecInterface.model1.DA = modbusHandlerGetSlaveAddress();
}

void setSUNSPEC_MODEL1_DA (void) {
    modbusHandlerSetSlaveAddress(sunspecInterface.model1.DA);
}
"""

def Type(name, type):
    return pycparser.c_ast.TypeDecl(
        declname=name,
        quals=[],
        type=pycparser.c_ast.IdentifierType(
            names=(type,),
        ),
    )

Decl = functools.partial(
    pycparser.c_ast.Decl,
    name=None,
    quals=[],
    storage=[],
    funcspec=[],
    init=None,
    bitsize=None,
)

TypeDecl = functools.partial(
    pycparser.c_ast.TypeDecl,
    declname='',
    quals=[],
)


def function(name):
    type_decl = Type(
        name=name,
        type=['void'],
    )

    param_list = pycparser.c_ast.ParamList(
        params=[
            pycparser.c_ast.Typename(
                name=None,
                quals=[],
                type=pycparser.c_ast.IdentifierType(
                    names=['void'],
                ),
            ),
        ],
    )

    func_decl = pycparser.c_ast.FuncDecl(
        args=param_list,
        type=type_decl,
    )

    decl = Decl(
        name=name,
        type=func_decl,
    )

    compound = pycparser.c_ast.Compound(
        block_items=[],
    )

    function = pycparser.c_ast.FuncDef(
        decl=decl,
        param_decls=None,
        body=compound,
    )

    return function


def test_():
    ast = parse(c)
    ast.ext.append(function('myFunctionName'))
    ast.show()



import epcpm.importexportdialog


def test_blank(qtbot):
    dialog = epcpm.importexportdialog.Dialog()

    with qtbot.waitSignal(dialog.accepted):
        with qtbot.waitExposed(dialog):
            dialog.open()

        dialog.accept()

    assert dialog.paths_result == epcpm.importexportdialog.ImportPaths(
        can=None,
        hierarchy=None,
        tables_c=None,
        spreadsheet=None,
        spreadsheet_user=None,
        smdx=[],
        sunspec_tables_c=None,
        sunspec_c=None,
        sil_c=None,
        interface_c=None,
    )


def test_cancel(qtbot):
    dialog = epcpm.importexportdialog.Dialog()

    with qtbot.waitSignal(dialog.rejected):
        with qtbot.waitExposed(dialog):
            dialog.open()

        dialog.reject()

    assert dialog.paths_result is None



import attr
import epyqlib.utils.general
import epyqlib.pm.parametermodel

import epcpm.cantotablesc
import epcpm.sunspecmodel


builders = epyqlib.utils.general.TypeMap()


def export(c_path, h_path, sunspec_model, skip_sunspec=False):
    builder = builders.wrap(
        wrapped=sunspec_model.root,
        parameter_uuid_finder=sunspec_model.node_from_uuid,
        skip_sunspec=skip_sunspec,
    )

    c_path.parent.mkdir(parents=True, exist_ok=True)
    c_content, h_content = builder.gen()

    template_context = {
        'definitions': c_content,
        'declarations': h_content,
    }

    epcpm.c.render(
        source=c_path.with_suffix(f'{c_path.suffix}_pm'),
        destination=c_path,
        context=template_context,
    )

    epcpm.c.render(
        source=h_path.with_suffix(f'{h_path.suffix}_pm'),
        destination=h_path,
        context=template_context,
    )


@builders(epcpm.sunspecmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    skip_sunspec = attr.ib(default=False)

    def gen(self):
        both_lines = [[], []]

        # table_results = []

        if not self.skip_sunspec:
            for child in self.wrapped.children:
                if not isinstance(child, epcpm.sunspecmodel.Model):
                    continue

                builder = builders.wrap(
                    wrapped=child,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                )

                # table_results.append(builder.gen())
                # lines.extend(table_results[-1].table_lines)

                for lines, more_lines in zip(both_lines, builder.gen()):
                    lines.extend(more_lines)
                    lines.append('')

            # active_curves = parameter_query.child_by_name('ActiveCurves')
            #
            # for get_or_set in ('get', 'set'):
            #     active_lines = []
            #     for table_result in table_results:
            #         active_lines.extend(
            #             table_result.active_curves_lines[get_or_set],
            #         )
            #
            #     lines.extend([
            #         f'void {get_or_set}{active_curves.name}(void)',
            #         '{',
            #         active_lines,
            #         '}',
            #         '',
            #     ])

        return tuple(
            epyqlib.utils.general.format_nested_lists(lines)
            for lines in both_lines
        )


@builders(epcpm.sunspecmodel.Model)
@attr.s
class Model:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        for child in self.wrapped.children:
            found = isinstance(
                child,
                epcpm.sunspecmodel.TableRepeatingBlockReference,
            )
            if found:
                break
        else:
            return [[], []]

        builder = builders.wrap(
            wrapped=child,
            parameter_uuid_finder=self.parameter_uuid_finder,
            model_id=self.wrapped.id,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.TableRepeatingBlockReference)
@attr.s
class TableRepeatingBlockReference:
    wrapped = attr.ib()
    model_id = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        builder = builders.wrap(
            wrapped=self.wrapped.original,
            model_id=self.model_id,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        return builder.gen()


@builders(epcpm.sunspecmodel.TableRepeatingBlock)
@attr.s
class TableRepeatingBlock:
    wrapped = attr.ib()
    model_id = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        both_lines = [[], []]

        curve_group_string = ''.join(
            self.parameter_uuid_finder(uuid).name
            for uuid in self.wrapped.path[:-1]
        )
        curve_type = epcpm.cantotablesc.get_curve_type(curve_group_string)

        for curve_index in range(self.wrapped.repeats):
            for point in self.wrapped.children:
                builder = builders.wrap(
                    wrapped=point,
                    model_id=self.model_id,
                    curve_index=curve_index,
                    curve_type=curve_type,
                    parameter_uuid_finder=self.parameter_uuid_finder,
                )

                for lines, more_lines in zip(both_lines, builder.gen()):
                    lines.extend(more_lines)
                    lines.append('')

        return both_lines


@builders(epcpm.sunspecmodel.DataPoint)
@attr.s
class DataPoint:
    wrapped = attr.ib()
    model_id = attr.ib()
    curve_index = attr.ib()
    curve_type = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        table_element = self.parameter_uuid_finder(self.wrapped.parameter_uuid)
        curve_parent = table_element.tree_parent.tree_parent.tree_parent
        table_element = curve_parent.descendent(
            str(self.curve_index),
            table_element.tree_parent.name,
            table_element.name,
        )

        parameter_table = self.parameter_uuid_finder(
            self.wrapped.tree_parent.tree_parent.parameter_table_uuid,
        )

        array_or_group_element = table_element.original

        is_a_parameter = isinstance(
            array_or_group_element,
            epyqlib.pm.parametermodel.Parameter,
        )
        if is_a_parameter:
            parameter = array_or_group_element
        else:
            parameter = array_or_group_element.original

        is_group = isinstance(
            array_or_group_element.tree_parent,
            epyqlib.pm.parametermodel.Group,
        )
        is_array = isinstance(
            array_or_group_element.tree_parent,
            epyqlib.pm.parametermodel.Array,
        )

        if is_group:
            getter_setter = {
                'get': array_or_group_element.can_getter,
                'set': array_or_group_element.can_setter,
            }
        elif is_array:
            getter_setter = {
                'get': parameter_table.can_getter,
                'set': parameter_table.can_setter,
            }

        if getter_setter['get'] is None:
            getter_setter['get'] = ''

        if getter_setter['set'] is None:
            getter_setter['set'] = ''

        axis = table_element.tree_parent.axis
        if axis is None:
            axis = '<no axis>'

        interface_variable = (
            f'sunspecInterface'
            f'.model{self.model_id}'
            f'.Curve_{self.curve_index + 1:02}_{table_element.abbreviation}'
        )

        both_lines = [[], []]

        for get_set, embedded in getter_setter.items():
            # TODO: CAMPid 075780541068182645821856068542023499
            converter = {
                'uint32': {
                    'get': 'sunspecUint32ToSSU32_returns',
                    'set': 'sunspecSSU32ToUint32',
                },
                'int32': {
                    # TODO: add this to embedded?
                    # 'get': 'sunspecInt32ToSSS32',
                    'set': 'sunspecSSS32ToInt32',
                },
                'uint64': {
                    'get': 'sunspecUint64ToSSU64_returns',
                    'set': 'sunspecSSU64ToUint64',
                },
            }.get(self.parameter_uuid_finder(self.wrapped.type_uuid).name)

            body_lines = []

            if parameter.uses_interface_item():
                # TODO: CAMPid 9685439641536675431653179671436
                item_uuid_string = str(table_element.uuid).replace('-', '_')
                item_name = f'interfaceItem_{item_uuid_string}'

                if is_group:
                    if get_set == 'get':
                        body_lines.extend([
                            f'{item_name}.common.sunspec.getter(',
                            [
                                f'(InterfaceItem_void *) &{item_name},',
                                f'Meta_Value',
                            ],
                            f');',
                        ])
                    elif get_set == 'set':
                        body_lines.extend([
                            f'{item_name}.common.sunspec.setter(',
                            [
                                f'(InterfaceItem_void *) &{item_name},',
                                f'true,',
                                f'Meta_Value',
                            ],
                            f');',
                        ])
                elif is_array:
                    if get_set == 'get':
                        body_lines.extend([
                            f'{item_name}.table_common->common.sunspec.getter(',
                            [
                                f'(InterfaceItem_void *) &{item_name},',
                                f'Meta_Value',
                            ],
                            f');',
                        ])
                    elif get_set == 'set':
                        body_lines.extend([
                            f'{item_name}.table_common->common.sunspec.setter(',
                            [
                                f'(InterfaceItem_void *) &{item_name},',
                                f'true,',
                                f'Meta_Value',
                            ],
                            f');',
                        ])
            elif converter is not None:
                converter = converter[get_set]
                if get_set == 'get':
                    formatted = embedded.format(
                        curve_type=self.curve_type,
                        interface_signal=interface_variable,
                        point_index=table_element.index,
                        axis=axis,
                        upper_axis=axis.upper(),
                        curve_index=self.curve_index,
                    )
                    left, equals, right = (
                        element.strip()
                        for element in formatted.partition('=')
                    )

                    right = right.rstrip(';')

                    if equals != '=':
                        raise Exception('do not yet know how to handle this')

                    body_lines.append(
                        f'{left} = {converter}({right});',
                    )
                elif get_set == 'set':
                    body_lines.append(embedded.format(
                        curve_type=self.curve_type,
                        interface_signal=f'{converter}(&{interface_variable})',
                        point_index=table_element.index,
                        axis=axis,
                        upper_axis=axis.upper(),
                        curve_index=self.curve_index,
                    ))
            else:
                body_lines.append(embedded.format(
                    curve_type=self.curve_type,
                    interface_signal=interface_variable,
                    point_index=table_element.index,
                    axis=axis,
                    upper_axis=axis.upper(),
                    curve_index=self.curve_index,
                ))

            function_name = '_'.join([
                f'{get_set}SunspecModel{self.model_id}',
                'Curve',
                f'{self.curve_index + 1:02}',
                table_element.abbreviation,
            ])
            function_signature = f'void {function_name} (void)'

            both_lines[0].extend([
                f'{function_signature} {{',
                body_lines,
                '}',
                '',
            ])

            both_lines[1].append(
                f'{function_signature};',
            )

        return both_lines



# This file has been generated


build_system = 'AppVeyor'
build_id = '29670643'
build_number = '559'
build_version = '1.0.559'
job_id = 'vkbtowmalruayk2s'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/pm/build/job/vkbtowmalruayk2s'



import pytest

import epcpm.symtoproject
import epyqlib.pm


@pytest.fixture
def access_levels():
    access_levels = epyqlib.pm.parametermodel.AccessLevels()
    user_level = epyqlib.pm.parametermodel.AccessLevel(name='User', value=0)
    factory_level = epyqlib.pm.parametermodel.AccessLevel(name='Factory', value=1)
    
    access_levels.append_child(user_level)
    access_levels.append_child(factory_level)

    return access_levels

@pytest.fixture
def variant_cfgs():
    variants = epyqlib.pm.parametermodel.Enumeration()
    enumerator_type = epyqlib.pm.parametermodel.Enumerator
    
    cfgs = ['None', 'MG3', 'MG4', 'DG', 'HY', 'DC']
    val = 0
    for cfg in cfgs:
        e = enumerator_type(name=cfg, value=val)
        variants.append_child(e)
        val=val+1

    variants = [
        variant 
        for variant in variants.children 
        if variant.name != 'None'
    ]

    return variants


def test_strip_access_level_factory(access_levels):
    comment = 'abc <factory> def'
    
    stripped, level = epcpm.symtoproject.strip_access_level(comment, access_levels)
    
    assert stripped.split() == ['abc', 'def']
    assert level == access_levels.by_name('factory')


def test_strip_access_level_normal(access_levels):
    comment = 'abc def'
    
    stripped, level = epcpm.symtoproject.strip_access_level(comment, access_levels)
    
    assert stripped == comment
    assert level == access_levels.by_name('user')


def test_strip_variant_parameter_specific(variant_cfgs):
    comment = 'abc <HY> <DG> <MG4> def'
      
    stripped, variants = epcpm.symtoproject.strip_variant_parameter_tag(comment, variant_cfgs)
    variant_names = [
        variant.name 
        for variant in variants
    ]
    assert stripped.split() == ['abc', 'def']
    assert variant_names == ['MG4', 'DG', 'HY']


def test_strip_variant_parameter_normal(variant_cfgs):
    comment = 'abc def'
      
    stripped, variants = epcpm.symtoproject.strip_variant_parameter_tag(comment, variant_cfgs)
    variant_names = [
        variant.name 
        for variant in variants
    ]
    assert stripped == comment
    assert variant_names == ['MG3', 'MG4', 'DG', 'HY', 'DC']



import decimal
import io
import json
import textwrap

import pytest

import epyqlib.pm.parametermodel
import epyqlib.tests.common

import epcpm.canmodel
import epcpm.symtoproject


@pytest.fixture
def sym_file():
    return io.BytesIO(textwrap.dedent('''\
    FormatVersion=5.0 // Do not edit this line!
    Title="canmatrix-Export"
    
    {ENUMS}
    enum AccessLevel(0="User", 1="Engineering", 2="Factory")
    enum CmmControlsVariant(0="None", 1="MG3", 2="MG4", 3="DG", 4="HY", 5="DC")
    
    {SEND}
    
    [ParameterQuery]
    ID=1DEFF741h
    Type=Extended
    DLC=8
    Mux=TestMux 0,14 0 
    Var=TestParam unsigned 14,2 /f:0.01  /min:0.01  /max:0.2 /p:2 /d:0.02
    Var=FactoryParam unsigned 14,2 /f:0.01  /min:0.01  /max:0.2 /p:2 /d:0.02  // before <factory> after
    
    {SENDRECEIVE}
    
    [ParameterResponse]
    ID=1DEF41F7h
    Type=Extended
    DLC=8
    Mux=TestMux 0,14 0 
    Var=TestParam unsigned 14,2 /f:0.01  /min:0.01  /max:0.2 /p:2 /d:0.02
    Var=FactoryParam unsigned 14,2 /f:0.01  /min:0.01  /max:0.2 /p:2 /d:0.02  // before <factory> after
    ''').encode('utf-8'))


other_group_name = 'Uncategorized Stuff'


@pytest.fixture
def hierarchy_file():
    return io.StringIO(json.dumps({
        'children': [
            {
                'name': 'Test Group',
                'children': [
                    [
                        "TestMux",
                        "TestParam"
                    ],
                    [
                        "TestMux",
                        "FactoryParam"
                    ]
                ]
            },
            {
                'name': other_group_name,
                'unreferenced': True,
            },
        ]
    }))


@pytest.fixture
def empty_hierarchy_file():
    return io.StringIO('{"children": []}')


def test_other_group_name(hierarchy_file):
    hierarchy = json.load(hierarchy_file)
    assert epcpm.symtoproject.get_other_name(hierarchy) == other_group_name


def test_load_can_file():
    parameter_root, can_root, sunspec_root = epcpm.symtoproject.load_can_path(
        epyqlib.tests.common.symbol_files['customer'],
        epyqlib.tests.common.hierarchy_files['customer'],
    )

    assert isinstance(parameter_root, epyqlib.pm.parametermodel.Root)
    assert isinstance(can_root, epcpm.canmodel.Root)
    assert isinstance(sunspec_root, epcpm.sunspecmodel.Root)


def test_only_one_parameter_per_query_response_pair(
        sym_file,
        empty_hierarchy_file,
):
    parameter_root, *_ = epcpm.symtoproject.load_can_file(
        can_file=sym_file,
        file_type='sym',
        parameter_hierarchy_file=empty_hierarchy_file,
    )

    parameters = next(
        node
        for node in parameter_root.children
        if node.name == 'Parameters'
    )

    assert len(parameters.children[0].children) == 2


def test_access_level(sym_file, hierarchy_file):
    parameter_root, *_ = epcpm.symtoproject.load_can_file(
        can_file=sym_file,
        file_type='sym',
        parameter_hierarchy_file=hierarchy_file,
    )

    access_levels, = parameter_root.nodes_by_attribute(
        attribute_name='name',
        attribute_value='AccessLevel',
    )

    regular_parameter, = parameter_root.nodes_by_attribute(
        attribute_name='name',
        attribute_value='TestParam',
    )

    assert regular_parameter.access_level_uuid == access_levels.default().uuid

    factory_parameter, = parameter_root.nodes_by_attribute(
        attribute_name='name',
        attribute_value='FactoryParam',
    )

    assert (
        factory_parameter.access_level_uuid
        == access_levels.by_name('factory').uuid
    )
    assert 'before  after' == factory_parameter.comment


def test_accurate_decimal(sym_file, hierarchy_file):
    parameter_root, *_ = epcpm.symtoproject.load_can_file(
        can_file=sym_file,
        file_type='sym',
        parameter_hierarchy_file=hierarchy_file,
    )

    parameters = next(
        node
        for node in parameter_root.children
        if node.name == 'Parameters'
    )

    test_group = next(
        node
        for node in parameters.children
        if node.name == 'Test Group'
    )

    test_parameter = next(
        node
        for node in test_group.children
        if node.name.endswith('TestParam')
    )

    # TODO: default is probably in wrong scaling
    # assert isinstance(test_parameter.default, decimal.Decimal)
    # assert test_parameter.default == decimal.Decimal('0.02')

    assert isinstance(test_parameter.minimum, decimal.Decimal)
    assert test_parameter.minimum == decimal.Decimal('0.01')



import itertools

import attr
import docx
import docx.enum.section
import docx.enum.text

import epyqlib.cangenmanual
import epyqlib.pm.parametermodel
import epyqlib.treenode
import epyqlib.utils.general

builders = epyqlib.utils.general.TypeMap()


@attr.s
class Row:
    name = attr.ib()
    indent = attr.ib(default=0)
    fill = attr.ib(default='')
    factor = attr.ib(default='')
    units = attr.ib(default='')
    default = attr.ib(default='')
    minimum = attr.ib(default='')
    maximum = attr.ib(default='')
    enumeration = attr.ib(default='')
    comment = attr.ib(default='')

    def to_tuple(self, max_indent=5):
        return (
            *((self.fill,) * self.indent),
            self.name,
            *((self.fill,) * (max_indent - self.indent - 1)),
            self.factor,
            self.units,
            self.default,
            self.minimum,
            self.maximum,
            self.enumeration,
            self.comment,
        )


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    can_root = attr.ib()
    template = attr.ib()
    access_level = attr.ib()

    def gen(self):
        headings = Row(
            name='Name',
            factor='Factor',
            units='Units',
            default='Default',
            minimum='Min',
            maximum='Max',
            enumeration='Enumeration',
            comment='Comment',
        )
        max_indent = 5
        fill_width = 0.25
        name_width = 2.75
        widths = Row(
            name=name_width - (fill_width * (max_indent - 1)),
            indent=max_indent - 1,
            fill=fill_width,
            factor=0.625,
            units=0.625,
            default=0.875,
            minimum=0.625,
            maximum=0.625,
            enumeration=1.5,
            comment=None,
        )

        for heading, width in zip(headings.to_tuple(), widths.to_tuple()):
            print(heading, width)

        table = epyqlib.cangenmanual.Table(
            title=None,
            headings=headings.to_tuple(),
            widths=widths.to_tuple(),
            total_width=10,
        )

        import time
        start = time.monotonic()

        for child in self.wrapped.children:
            if child.name.endswith('Other'):
                continue

            try:
                builder = builders.wrap(
                    wrapped=child,
                    parameter_root=self.wrapped,
                    can_root=self.can_root,
                    access_level=self.access_level,
                )
            except KeyError:
                continue

            table.rows.extend(builder.gen(indent=0))

        now = time.monotonic()
        delta = now - start
        start = now
        print('rows built', int(delta))

        raw_rows = table.rows

        table.rows = tuple(
            row.to_tuple(max_indent=max_indent)
            for row in table.rows
        )

        now = time.monotonic()
        delta = now - start
        start = now
        print('rows converted', int(delta))

        if self.template is not None:
            doc = docx.Document(self.template)
        else:
            doc = docx.Document()

        doc_table = doc.add_table(rows=0, cols=len(table.headings))
        doc_table.autofit = False

        table.fill_docx(doc_table)

        epyqlib.cangenmanual.set_repeat_table_header(doc_table.rows[0])
        epyqlib.cangenmanual.prevent_row_breaks(doc_table)

        for row, doc_row in zip(itertools.chain((headings,), raw_rows), doc_table.rows):
            base_cell = doc_row.cells[row.indent]
            base_cell.merge(doc_row.cells[max_indent - 1])

        for row in doc_table.rows:
            for cell in row.cells:
                cell.text = cell.text.strip()
                cell.paragraphs[0].paragraph_format.line_spacing_rule = (
                    docx.enum.text.WD_LINE_SPACING.SINGLE
                )

        now = time.monotonic()
        delta = now - start
        start = now
        print('table filled', int(delta))

        for section in doc.sections:
            section.orientation = docx.enum.section.WD_ORIENTATION.LANDSCAPE

        return doc


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()
    parameter_root = attr.ib()
    can_root = attr.ib()
    access_level = attr.ib()

    def gen(self, indent):
        rows = [
            Row(
                name=self.wrapped.name,
                indent=indent,
            ),
        ]

        for child in self.wrapped.children:
            builder = builders.wrap(
                wrapped=child,
                parameter_root=self.parameter_root,
                can_root=self.can_root,
                access_level=self.access_level,
            )
            rows.extend(builder.gen(indent=indent + 1))

        return rows


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()
    parameter_root = attr.ib()
    can_root = attr.ib()
    access_level = attr.ib()

    def gen(self, indent):
        try:
            access_level = self.parameter_root.nodes_by_attribute(
                attribute_value=self.wrapped.access_level_uuid,
                attribute_name='uuid',
            ).pop()
        except epyqlib.treenode.NotFoundError:
            pass
        else:
            if access_level.value > self.access_level.value:
                print('skipping', self.wrapped.name)
                return []

        signal = self.can_root.nodes_by_attribute(
            attribute_value=self.wrapped.uuid,
            attribute_name='parameter_uuid',
        ).pop()

        factor = signal.factor
        if factor is None or factor == 1:
            factor = ''

        units = self.wrapped.units
        if units is None:
            units = ''

        try:
            enumeration = self.parameter_root.nodes_by_attribute(
                attribute_value=self.wrapped.enumeration_uuid,
                attribute_name='uuid',
            ).pop().name
        except epyqlib.treenode.NotFoundError:
            enumeration = ''

        default = self.wrapped.default
        if default is None:
            default = ''

        minimum = self.wrapped.minimum
        if minimum is None:
            minimum = ''

        maximum = self.wrapped.maximum
        if maximum is None:
            maximum = ''

        comment = self.wrapped.comment
        if comment is None:
            comment = ''

        return [
            Row(
                name=self.wrapped.name,
                indent=indent,
                factor=factor,
                units=units,
                default=default,
                minimum=minimum,
                maximum=maximum,
                enumeration=enumeration,
                comment=comment,
            ),
        ]



import pathlib
import textwrap

import click
import click.testing

import epcpm.cli.parameterstoc


parameters_path = (
    pathlib.Path(__file__).parents[0] / 'test_parameterstoc_parameters.json'
)


def test_declaration():
    runner = click.testing.CliRunner()
    result = runner.invoke(
        epcpm.cli.parameterstoc.cli,
        [
            '--parameters', parameters_path,
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
    struct GreenType_s
    {
      RedType red;
    };
    typedef struct GreenType_s GreenType_t;
    ''')


def test_instantiation():
    runner = click.testing.CliRunner()
    result = runner.invoke(
        epcpm.cli.parameterstoc.cli,
        [
            '--parameters', parameters_path,
            '--instantiation',
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert result.output == textwrap.dedent('''\
    GreenType_t green;
    BlueType blue;
    ''')



import os

import jinja2


# TODO: CAMPid 073407143081341008467657184603164130
def format_nested_lists(it, indent=''):
    result = []

    for item in it:
        if isinstance(item, list):
            result.extend(format_nested_lists(item, indent=indent + '    '))
        elif item.strip() == '':
            result.append('')
        else:
            result.append(indent + item)

    if indent == '':
        result.append('')
        return '\n'.join(result)
    else:
        return result


def render(source, destination, context={}, encoding='utf-8', newline='\n'):
    environment = jinja2.Environment(
        undefined=jinja2.StrictUndefined,
        loader=jinja2.FileSystemLoader(os.fspath(source.parent)),
        newline_sequence=newline,
        autoescape=False,
        trim_blocks=True,
    )
    template = environment.get_template(name=source.name)

    rendered = template.render(context)
    rendered = rendered.rstrip() + newline

    destination.write_bytes(rendered.encode(encoding))



# This file was automatically generated by sub\epyqlib\epyqlib\cli\generateversion.py
__version_build_time__ = '2019-12-20T03:44:17.108793'

__version__ = '2019.12.5'
__sha__ = 'a8161404103220464cef0b4cd71c20a845edc46e'
__revision__ = 'a8161404103220464cef0b4cd71c20a845edc46e'



import epyqlib.pm.parametermodel

import epcpm.parameterstohierarchy
import epcpm.canmodel


def test_():
    parameter_root = epyqlib.pm.parametermodel.Root()

    parameters = epyqlib.pm.parametermodel.Group(
        name='Parameters',
        uuid='b95feb06-1f19-4557-8335-4a2287222c69',
    )
    parameter_root.append_child(parameters)

    group = epyqlib.pm.parametermodel.Group(
        name='Group A',
        uuid='c202c286-4eef-4baa-b088-4195c200d944',
    )
    parameters.append_child(group)

    parameter_aa = epyqlib.pm.parametermodel.Parameter(
        name='Parameter AA',
        uuid='b9c38efa-1f2f-458f-980d-bce82ea082fe',
    )
    group.append_child(parameter_aa)
    parameter_ab = epyqlib.pm.parametermodel.Parameter(
        name='Parameter AB',
        uuid='c9d12fb7-e282-4170-bf5d-c880145c2fc5',
    )
    group.append_child(parameter_ab)

    parameter_a = epyqlib.pm.parametermodel.Parameter(
        name='Parameter A',
        uuid='253233ea-3b00-47ec-9b91-623bb5c3dca4',
    )
    parameters.append_child(parameter_a)

    can_root = epcpm.canmodel.Root()

    multiplexed_message = epcpm.canmodel.MultiplexedMessage(
        name='ParameterQuery',
        uuid='3bb49d82-e7fb-4cc1-af8d-f8701fe8b371',
    )
    can_root.append_child(multiplexed_message)

    multiplexer_a = epcpm.canmodel.Multiplexer(
        name='Multiplexer_A',
        uuid='78ce435e-1ba8-424f-aec4-212bbba0c612',
    )
    multiplexed_message.append_child(multiplexer_a)

    signal_aa = epcpm.canmodel.Signal(
        name='Signal_PAA',
        parameter_uuid=parameter_aa.uuid,
        uuid='3c25a151-2ae0-4a8d-bc77-704123aa6547',
    )
    multiplexer_a.append_child(signal_aa)

    multiplexer_b = epcpm.canmodel.Multiplexer(
        name='Multiplexer_B',
        uuid='464d14d3-9bd6-45af-aac8-29cb87faff23',
    )
    multiplexed_message.append_child(multiplexer_b)

    signal_ab = epcpm.canmodel.Signal(
        name='Signal_PAB',
        parameter_uuid=parameter_ab.uuid,
        uuid='84d3eb7a-212c-4634-b46f-92e51b6147fd',
    )
    multiplexer_b.append_child(signal_ab)

    signal_a = epcpm.canmodel.Signal(
        name='Signal_PA',
        parameter_uuid=parameter_a.uuid,
        uuid='db7a51b2-db17-427e-9acc-45d6a988d3d4',
    )
    multiplexer_b.append_child(signal_a)

    expected = {
        'children': [
            {
                'name': 'Group A',
                'children': [
                    [
                        'Multiplexer_A',
                        'Signal_PAA',
                    ],
                    [
                        'Multiplexer_B',
                        'Signal_PAB',
                    ],
                ]
            },
            [
                'Multiplexer_B',
                'Signal_PA',
            ],
        ]
    }

    builder = epcpm.parameterstohierarchy.builders.wrap(
        wrapped=parameter_root,
        can_root=can_root,
    )

    assert builder.gen(json_output=False) == expected



import collections
import pathlib

import epyqlib.tests.test_attrsmodel

import epcpm.project
import epcpm.sunspecmodel


# See file COPYING in this source tree
__copyright__ = 'Copyright 2018, EPC Power Corp.'
__license__ = 'GPLv2+'


this = pathlib.Path(__file__).resolve()
here = this.parent


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epcpm.sunspecmodel.types,
    root_type=epcpm.sunspecmodel.Root,
    columns=epcpm.sunspecmodel.columns,
)


def test_model_has_header():
    project = epcpm.project.loadp(here/'project'/'project.pmp')

    parameter_model = project.models.parameters
    enumerations = parameter_model.list_selection_roots['enumerations']
    sunspec_types = epcpm.sunspecmodel.build_sunspec_types_enumeration()
    enumerations.append_child(sunspec_types)
    parameter_model.list_selection_roots['sunspec types'] = sunspec_types

    types = project.models.parameters.list_selection_roots['sunspec types']

    model = epcpm.sunspecmodel.Model()
    parameters = model.children[0].add_data_points(
        uint16_uuid=types.child_by_name('uint16').uuid,
        model_id=1,
    )

    header_block = model.children[0]

    assert isinstance(header_block, epcpm.sunspecmodel.HeaderBlock)

    assert len(header_block.children) == len(parameters)
    for point, parameter in zip(header_block.children, parameters):
        assert point.parameter_uuid == parameter.uuid

    assert parameters[0].abbreviation == 'ID'
    assert parameters[1].abbreviation == 'L'

    assert header_block.offset == 0

    fixed_block = model.children[1]

    assert fixed_block.offset == 2


# TODO: CAMPid 094329054780541680163054608431067542971349
def count_types(sequence):
    counts = collections.defaultdict(int)

    for element in sequence:
        counts[type(element)] += 1

    return counts


def test_table_update_unlinked():
    expected_counts = {
        epcpm.sunspecmodel.TableRepeatingBlock: 8,
        epcpm.sunspecmodel.DataPoint: 2,
    }

    project = epcpm.project.loadp(here/'project'/'project.pmp')
    sunspec_table, = project.models.sunspec.root.nodes_by_attribute(
        attribute_value='First Table',
        attribute_name='name',
    )
    parameter_table = project.models.parameters.node_from_uuid(
        sunspec_table.parameter_table_uuid,
    )
    parameter_table.update()

    assert count_types(sunspec_table.children) == expected_counts

    sunspec_table.update()
    assert count_types(sunspec_table.children) == expected_counts

    print()
    print(epyqlib.utils.qt.indented_text_from_model(project.models.sunspec.model))

    sunspec_table.parameter_table_uuid = None
    sunspec_table.update()

    assert count_types(sunspec_table.children) == {}



import itertools
import math
import os
import pathlib
import subprocess

import attr
import graham

import epcpm.cantosym
import epcpm.cantotablesc
import epcpm.parameterstohierarchy
import epcpm.parameterstointerface
import epcpm.parameterstosil
import epcpm.project
import epcpm.smdxtosunspec
import epcpm.sunspecmodel
import epcpm.sunspecmodel
import epcpm.sunspectotablesc
import epcpm.sunspectomanualc
import epcpm.sunspectomanualh
import epcpm.sunspectoxlsx
import epcpm.symtoproject
import epyqlib.attrsmodel


def full_import(paths):
    with open(paths.can, 'rb') as sym, open(paths.hierarchy) as hierarchy:
        parameters_root, can_root, sunspec_root = (
            epcpm.symtoproject.load_can_file(
                can_file=sym,
                file_type=str(pathlib.Path(sym.name).suffix[1:]),
                parameter_hierarchy_file=hierarchy,
            )
        )

    project = epcpm.project.Project()

    project.models.parameters = epyqlib.attrsmodel.Model(
        root=parameters_root,
        columns=epyqlib.pm.parametermodel.columns,
    )
    project.models.can = epyqlib.attrsmodel.Model(
        root=can_root,
        columns=epcpm.canmodel.columns,
    )
    project.models.sunspec = epyqlib.attrsmodel.Model(
        root=sunspec_root,
        columns=epcpm.sunspecmodel.columns,
    )

    epcpm.project._post_load(project)

    # TODO: backmatching
    epcpm.symtoproject.go_add_tables(
        parameters_root=project.models.parameters.root,
        can_root=project.models.can.root,
    )

    sunspec_types = epcpm.sunspecmodel.build_sunspec_types_enumeration()
    enumerations = (
        project.models.parameters.list_selection_roots['enumerations']
    )
    enumerations.append_child(sunspec_types)

    project.models.update_enumeration_roots()

    sunspec_models = []
    prefix = 'smdx_'
    suffix = '.xml'
    for smdx_path in paths.smdx:
        models = epcpm.smdxtosunspec.import_models(
            int(smdx_path.name[len(prefix):-len(suffix)]),
            parameter_model=project.models.parameters,
            paths=[smdx_path.parent],
        )
        sunspec_models.extend(models)

    for sunspec_model in sunspec_models:
        project.models.sunspec.root.append_child(sunspec_model)

    points = (
        (model, block, point)
        for model in project.models.sunspec.root.children
        for block in model.children
        for point in block.children
    )

    get_set = epcpm.smdxtosunspec.import_get_set(paths.spreadsheet)

    for model, block, point in points:
        parameter = project.models.sunspec.node_from_uuid(
            point.parameter_uuid,
        )
        for direction in ('get', 'set'):
            key = epcpm.smdxtosunspec.GetSetKey(
                model=model.id,
                name=parameter.abbreviation,
                get_set=direction,
            )
            accessor = get_set.get(key)
            if accessor is not None:
                setattr(point, direction, accessor)

    project.paths['parameters'] = 'parameters.json'
    project.paths['can'] = 'can.json'
    project.paths['sunspec'] = 'sunspec.json'

    return project


def full_export(
        project,
        paths,
        target_directory,
        first_time=False,
        skip_sunspec=False,
        include_uuid_in_item=False,
):
    epcpm.cantosym.export(
        path=paths.can,
        can_model=project.models.can,
        parameters_model=project.models.parameters,
    )

    epcpm.parameterstohierarchy.export(
        path=paths.hierarchy,
        can_model=project.models.can,
        parameters_model=project.models.parameters,
    )

    epcpm.parameterstointerface.export(
        c_path=paths.interface_c,
        h_path=paths.interface_c.with_suffix('.h'),
        can_model=project.models.can,
        sunspec_model=project.models.sunspec,
        parameters_model=project.models.parameters,
        skip_sunspec=skip_sunspec,
        include_uuid_in_item=include_uuid_in_item,
    )

    epcpm.sunspectoxlsx.export(
        path=paths.spreadsheet,
        sunspec_model=project.models.sunspec,
        parameters_model=project.models.parameters,
        skip_sunspec=skip_sunspec,
    )

    epcpm.sunspectoxlsx.export(
        path=paths.spreadsheet_user,
        sunspec_model=project.models.sunspec,
        parameters_model=project.models.parameters,
        skip_sunspec=skip_sunspec,
        column_filter=attr.evolve(
            epcpm.sunspectoxlsx.attr_fill(epcpm.sunspectoxlsx.Fields, True), 
            get=False,
            set=False,
            item=False,
        )
    )

    epcpm.cantotablesc.export(
        path=paths.tables_c,
        can_model=project.models.can,
    )

    epcpm.sunspectotablesc.export(
        c_path=paths.sunspec_tables_c,
        h_path=paths.sunspec_tables_c.with_suffix('.h'),
        sunspec_model=project.models.sunspec,
        skip_sunspec=skip_sunspec,
    )

    epcpm.parameterstosil.export(
        c_path=paths.sil_c,
        h_path=paths.sil_c.with_suffix('.h'),
        parameters_model=project.models.parameters,
    )

    if first_time and not skip_sunspec:
        epcpm.sunspectomanualc.export(
            path=paths.sunspec_c,
            sunspec_model=project.models.sunspec,
        )

        epcpm.sunspectomanualh.export(
            path=paths.sunspec_c,
            sunspec_model=project.models.sunspec,
        )

    run_generation_scripts(target_directory, skip_sunspec=skip_sunspec)


def run_generation_scripts(base_path, skip_sunspec=False):
    scripts = base_path/'venv'/'Scripts'
    interface = base_path/'interface'

    subprocess.run(
        [
            os.fspath(scripts/'generatestripcollect'),
            os.fspath(interface/'EPC_DG_ID247_FACTORY.sym'),
            '-o',
            os.fspath(interface/'EPC_DG_ID247.sym'),
            '--hierarchy',
            os.fspath(interface/'EPC_DG_ID247_FACTORY.parameters.json'),
            '--hierarchy-out',
            os.fspath(interface/'EPC_DG_ID247.parameters.json'),
            '--device-file',
            os.fspath(interface/'devices.json'),
            '--output-directory',
            os.fspath(interface/'devices'),
        ],
        check=True,
    )

    emb_lib = base_path/'embedded-library'
    subprocess.run(
        [
            os.fspath(scripts/'sunspecparser'),
            os.fspath(emb_lib/'MODBUS_SunSpec-EPC.xlsx'),
        ],
        check=True,
    )


def modification_time_or(path, alternative):
    try:
        return path.stat().st_mtime
    except FileNotFoundError:
        return alternative


def get_sunspec_models(path):
    root_schema = graham.schema(epcpm.sunspecmodel.Root)
    raw = path.read_bytes()
    root = root_schema.loads(raw).data

    return tuple(
        child.id
        for child in root.children
        if isinstance(child, epcpm.sunspecmodel.Model)
    )


def is_stale(project, paths, skip_sunspec=False):
    loaded_project = epcpm.project.loadp(project, post_load=False)

    source_paths = (
        project,
        *(
            project.parent / path
            for path in attr.astuple(loaded_project.paths)
        ),
    )

    source_modification_time = max(
        path.stat().st_mtime
        for path in source_paths
    )

    if skip_sunspec:
        sunspec_models = []
    else:
        sunspec_models = get_sunspec_models(
            project.parent / loaded_project.paths.sunspec,
        )

    smdx = tuple(
        paths.sunspec_c/f'smdx_{model:05}.xml'
        for model in sunspec_models
    )

    sunspec_c_h = tuple(
        paths.sunspec_c/f'sunspecInterfaceGen{model}.{extension}'
        for model, extension in itertools.product(sunspec_models, ('c', 'h'))
    )

    sil_c_h = (
        paths.sil_c,
        paths.sil_c.with_suffix('.h')
    )

    destination_paths = [
        paths.can,
        paths.hierarchy,
        *paths.smdx,
        paths.spreadsheet,
        paths.spreadsheet_user,
        *smdx,
        *sunspec_c_h,
        paths.tables_c,
        *sil_c_h,
    ]

    destination_modification_time = min(
        modification_time_or(path=path, alternative=-math.inf)
        for path in destination_paths
    )

    destination_newer_by = (
            destination_modification_time - source_modification_time
    )

    return destination_newer_by < 1



import decimal
import os
import string
import re

import attr
import jinja2
import toolz

import epyqlib.pm.parametermodel
import epyqlib.utils.general

import epcpm.cantosym
import epcpm.sunspecmodel
import epcpm.sunspectoxlsx

builders = epyqlib.utils.general.TypeMap()


# TODO: move this somewhere common in python code...
sunspec_types = {
    'uint16': 'sunsU16',
    'enum16': 'sunsU16',
    'int16': 'sunsS16',
    'uint32': 'sunsU32',
    'int32': 'sunsS32',
}


class InvalidAccessLevelError(Exception):
    @classmethod
    def build(cls, value, parameter):
        message = (
            f'Invalid access level specified for'
            f' {parameter.name} ({parameter.uuid}): {value}'
        )

        return cls(message)


def export(
        c_path,
        h_path,
        parameters_model,
        can_model,
        sunspec_model,
        skip_sunspec=False,
        include_uuid_in_item=False,
):
    if skip_sunspec:
        sunspec_root = None
    else:
        sunspec_root = sunspec_model.root

    builder = builders.wrap(
        wrapped=parameters_model.root,
        can_root=can_model.root,
        sunspec_root=sunspec_root,
        include_uuid_in_item=include_uuid_in_item,
    )

    if skip_sunspec:
        model_ids = []
    else:
        model_ids = [
            1,
            17,
            103,
            120,
            121,
            122,
            123,
            126,
            129,
            130,
            132,
            134,
            135,
            136,
            137,
            138,
            141,
            142,
            145,
            160,
            *range(65000, 65011),
            65534,
        ]

    c_path.parent.mkdir(parents=True, exist_ok=True)

    built_c, built_h = builder.gen()

    template_context = {
        'sunspec_interface_gen_headers': (
            f'sunspecInterfaceGen{id}.h'
            for id in model_ids
        ),
        'sunspec_interface_headers': (
            f'sunspecInterface{id:05}.h'
            for id in model_ids
        ),
        'interface_items': epcpm.c.format_nested_lists(
            built_c,
        ).strip(),
        'declarations': epcpm.c.format_nested_lists(built_h).strip(),
    }

    epcpm.c.render(
        source=c_path.with_suffix(f'{c_path.suffix}_pm'),
        destination=c_path,
        context=template_context,
    )

    epcpm.c.render(
        source=h_path.with_suffix(f'{h_path.suffix}_pm'),
        destination=h_path,
        context=template_context,
    )


@builders(epyqlib.pm.parametermodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()

    def gen(self):
        parameters = next(
            node
            for node in self.wrapped.children
            if node.name == 'Parameters'
        )

        def can_node_wanted(node):
            if getattr(node, 'parameter_uuid', None) is None:
                return False

            parameter_query_parent = node.tree_parent.tree_parent

            is_a_can_table = isinstance(
                node.tree_parent.tree_parent,
                epcpm.canmodel.CanTable,
            )
            if is_a_can_table:
                parameter_query_parent = parameter_query_parent.tree_parent

            is_a_query = (
                getattr(parameter_query_parent, 'name', '')
                == 'ParameterQuery'
            )
            if not is_a_query:
                return False

            return True

        can_nodes_with_parameter_uuid = self.can_root.nodes_by_filter(
            filter=can_node_wanted,
        )

        parameter_uuid_to_can_node = {
            node.parameter_uuid: node
            for node in can_nodes_with_parameter_uuid
        }

        def sunspec_node_wanted(node):
            if getattr(node, 'parameter_uuid', None) is None:
                return False

            if not isinstance(node, epcpm.sunspecmodel.DataPoint):
                return False

            return True

        if self.sunspec_root is None:
            parameter_uuid_to_sunspec_node = {}
        else:
            sunspec_nodes_with_parameter_uuid = self.sunspec_root.nodes_by_filter(
                filter=sunspec_node_wanted,
            )

            parameter_uuid_to_sunspec_node = {
                node.parameter_uuid: node
                for node in sunspec_nodes_with_parameter_uuid
            }

        lengths_equal = (
            len(can_nodes_with_parameter_uuid)
            == len(parameter_uuid_to_can_node)
        )
        if not lengths_equal:
            raise Exception()

        c = []
        h = []

        for child in parameters.children:
            if not isinstance(
                    child,
                    (
                        epyqlib.pm.parametermodel.Group,
                        epyqlib.pm.parametermodel.Parameter,
                        epyqlib.pm.parametermodel.Table,
                        # epcpm.parametermodel.EnumeratedParameter,
                    ),
            ):
                continue

            c_built, h_built = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                include_uuid_in_item=self.include_uuid_in_item,
                parameter_uuid_to_can_node=parameter_uuid_to_can_node,
                parameter_uuid_to_sunspec_node=(
                    parameter_uuid_to_sunspec_node
                ),
                parameter_uuid_finder=self.wrapped.model.node_from_uuid,
            ).gen()

            c.extend(c_built)
            h.extend(h_built)

        return c, h

        # return itertools.chain.from_iterable(
        #     builders.wrap(
        #         wrapped=child,
        #         can_root=self.can_root,
        #         sunspec_root=self.sunspec_root,
        #         parameter_uuid_to_can_node=parameter_uuid_to_can_node,
        #         parameter_uuid_to_sunspec_node=(
        #             parameter_uuid_to_sunspec_node
        #         ),
        #         parameter_uuid_finder=self.wrapped.model.node_from_uuid,
        #     ).gen()
        #     for child in parameters.children
        #     if isinstance(
        #         child,
        #         (
        #             epyqlib.pm.parametermodel.Group,
        #             epyqlib.pm.parametermodel.Parameter,
        #             # epcpm.parametermodel.EnumeratedParameter,
        #         ),
        #     )
        # )


@builders(epyqlib.pm.parametermodel.Group)
@attr.s
class Group:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        c = []
        h = []
        for child in self.wrapped.children:
            if not isinstance(
                    child,
                    (
                        epyqlib.pm.parametermodel.Group,
                        epyqlib.pm.parametermodel.Parameter,
                        epyqlib.pm.parametermodel.Table,
                        # epcpm.parametermodel.EnumeratedParameter,
                    ),
            ):
                continue

            c_built, h_built = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                include_uuid_in_item=self.include_uuid_in_item,
                parameter_uuid_to_can_node=(
                    self.parameter_uuid_to_can_node
                ),
                parameter_uuid_to_sunspec_node=(
                    self.parameter_uuid_to_sunspec_node
                ),
                parameter_uuid_finder=self.parameter_uuid_finder,
            ).gen()

            c.extend(c_built)
            h.extend(h_built)

        return c, h
        # return itertools.chain.from_iterable(
        #     result
        #     for result in (
        #         builders.wrap(
        #             wrapped=child,
        #             can_root=self.can_root,
        #             sunspec_root=self.sunspec_root,
        #             parameter_uuid_to_can_node=(
        #                 self.parameter_uuid_to_can_node
        #             ),
        #             parameter_uuid_to_sunspec_node=(
        #                 self.parameter_uuid_to_sunspec_node
        #             ),
        #             parameter_uuid_finder=self.parameter_uuid_finder,
        #         ).gen()
        #         for child in self.wrapped.children
        #         if isinstance(
        #             child,
        #             (
        #                 epyqlib.pm.parametermodel.Group,
        #                 epyqlib.pm.parametermodel.Parameter,
        #                 # epcpm.parametermodel.EnumeratedParameter,
        #             ),
        #         )
        #     )
        # )


@builders(epyqlib.pm.parametermodel.Parameter)
@attr.s
class Parameter:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        parameter = self.wrapped
        can_signal = self.parameter_uuid_to_can_node.get(parameter.uuid)
        sunspec_point = self.parameter_uuid_to_sunspec_node.get(parameter.uuid)

        interface_data = [
            can_signal,
            sunspec_point,
        ]

        uses_interface_item = (
            isinstance(parameter, epyqlib.pm.parametermodel.Parameter)
            and parameter.uses_interface_item()
        )

        if not uses_interface_item or all(x is None for x in interface_data):
            return [[], []]

        scale_factor_variable = 'NULL'
        scale_factor_updater = 'NULL'

        if parameter.getter_function is None:
            getter_function = 'NULL'
        else:
            getter_function = parameter.getter_function

        if parameter.setter_function is None:
            setter_function = 'NULL'
        else:
            setter_function = parameter.setter_function

        if parameter.internal_variable is not None:
            var_or_func = 'variable'

            variable_or_getter_setter = [
                f'.variable = &{parameter.internal_variable},',
            ]
        else:
            var_or_func = 'functions'

            variable_or_getter_setter = [
                f'.getter = {getter_function},',
            ]

        variable_or_getter_setter.append(f'.setter = {setter_function},')

        if sunspec_point is None:
            sunspec_variable = 'NULL'
            sunspec_getter = 'NULL'
            sunspec_setter = 'NULL'
            hand_coded_sunspec_getter_function = 'NULL'
            hand_coded_sunspec_setter_function = 'NULL'
        else:
            model_id = sunspec_point.tree_parent.tree_parent.id
            # TODO: move this somewhere common in python code...
            sunspec_type = sunspec_types[
                self.parameter_uuid_finder(sunspec_point.type_uuid).name
            ]

            # TODO: handle tables with repeating blocks and references

            hand_coded_getter_function_name = epcpm.sunspectoxlsx.getter_name(
                parameter=parameter,
                model_id=model_id,
                is_table=False,
            )

            hand_coded_setter_function_name = epcpm.sunspectoxlsx.setter_name(
                parameter=parameter,
                model_id=model_id,
                is_table=False,
            )

            if sunspec_point.hand_coded_getter:
                hand_coded_sunspec_getter_function = (
                    f'&{hand_coded_getter_function_name}'
                )
            else:
                hand_coded_sunspec_getter_function = 'NULL'

            if sunspec_point.hand_coded_setter:
                hand_coded_sunspec_setter_function = (
                    f'&{hand_coded_setter_function_name}'
                )
            else:
                hand_coded_sunspec_setter_function = 'NULL'

            sunspec_model_variable = f'sunspecInterface.model{model_id}'

            # TODO: CAMPid 67549654267913467967436
            if sunspec_point.factor_uuid is not None:
                factor_point = self.sunspec_root.model.node_from_uuid(
                    sunspec_point.factor_uuid,
                )
                sunspec_scale_factor = self.parameter_uuid_finder(
                    factor_point.parameter_uuid,
                ).abbreviation

                scale_factor_variable = (
                    f'&{sunspec_model_variable}.{sunspec_scale_factor}'
                )
                scale_factor_updater_name = (
                    f'getSUNSPEC_MODEL{model_id}_{sunspec_scale_factor}'
                )
                scale_factor_updater = f'&{scale_factor_updater_name}'

            sunspec_variable = (
                f'&{sunspec_model_variable}.{parameter.abbreviation}'
            )

            # TODO: CAMPid 9675436715674367943196954756419543975314
            getter_setter_list = [
                'InterfaceItem',
                var_or_func,
                parameter.internal_type,
                sunspec_type,
            ]

            sunspec_getter = '_'.join(
                str(x)
                for x in getter_setter_list + ['getter']
            )
            sunspec_setter = '_'.join(
                str(x)
                for x in getter_setter_list + ['setter']
            )

        interface_item_type = (
            f'InterfaceItem_{var_or_func}_{parameter.internal_type}'
        )

        can_getter, can_setter, can_variable = can_getter_setter_variable(
            can_signal=can_signal,
            parameter=parameter,
            var_or_func_or_table=var_or_func,
        )

        access_level = get_access_level_string(
            parameter=parameter,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        result = create_item(
            item_uuid=parameter.uuid,
            include_uuid_in_item=self.include_uuid_in_item,
            access_level=access_level,
            can_getter=can_getter,
            can_setter=can_setter,
            can_variable=can_variable,
            hand_coded_sunspec_getter_function=hand_coded_sunspec_getter_function,
            hand_coded_sunspec_setter_function=hand_coded_sunspec_setter_function,
            interface_item_type=interface_item_type,
            internal_scale=parameter.internal_scale_factor,
            meta_initializer_values=create_meta_initializer_values(parameter),
            parameter=parameter,
            scale_factor_updater=scale_factor_updater,
            scale_factor_variable=scale_factor_variable,
            sunspec_getter=sunspec_getter,
            sunspec_setter=sunspec_setter,
            sunspec_variable=sunspec_variable,
            variable_or_getter_setter=variable_or_getter_setter,
            can_scale_factor=getattr(can_signal, 'factor', None),
        )

        return result


@attr.s(frozen=True)
class FixedWidthType:
    name = attr.ib()
    bits = attr.ib()
    signed = attr.ib()
    minimum_code = attr.ib()
    maximum_code = attr.ib()

    @classmethod
    def build(cls, bits, signed):
        return cls(
            name=fixed_width_name(bits=bits, signed=signed),
            bits=bits,
            signed=signed,
            minimum_code=fixed_width_limit_text(
                bits=bits,
                signed=signed,
                limit='min',
            ),
            maximum_code=fixed_width_limit_text(
                bits=bits,
                signed=signed,
                limit='max',
            ),
        )


@attr.s(frozen=True)
class FloatingType:
    name = attr.ib()
    bits = attr.ib()
    minimum_code = attr.ib()
    maximum_code = attr.ib()

    @classmethod
    def build(cls, bits):
        return cls(
            name={32: 'float', 64: 'double'}[bits],
            bits=bits,
            minimum_code='(-INFINITY)',
            maximum_code='(INFINITY)',
        )


@attr.s(frozen=True)
class BooleanType:
    name = attr.ib(default='bool')
    bits = attr.ib(default=2)
    minimum_code = attr.ib(default='(false)')
    maximum_code = attr.ib(default='(true)')


@attr.s(frozen=True)
class SizeType:
    name = attr.ib(default='size_t')
    bits = attr.ib(default=32)
    minimum_code = attr.ib(default='(0)')
    maximum_code = attr.ib(default='(SIZE_MAX)')


def fixed_width_name(bits, signed):
    if signed:
        u = ''
    else:
        u = 'u'

    return f'{u}int{bits}_t'


def fixed_width_limit_text(bits, signed, limit):
    limits = ('min', 'max')

    if limit not in limits:
        raise Exception(f'Requested limit not found in {list(limits)}: {limit:!r}')

    if not signed and limit == 'min':
        return '(0U)'

    u = '' if signed else 'U'

    return f'({u}INT{bits}_{limit.upper()})'


types = {
    type.name: type
    for type in (
        *(
            FixedWidthType.build(
                bits=bits,
                signed=signed,
            )
            for bits in (8, 16, 32, 64)
            for signed in (False, True)
        ),
        *(
            FloatingType.build(bits=bits)
            for bits in (32, 64)
        ),
        BooleanType(),
        SizeType(),
    )
}


def create_meta_initializer_values(parameter):
    def create_literal(value, type):
        value *= decimal.Decimal(10) ** parameter.internal_scale_factor

        suffix = ''

        if type == 'float':
            suffix = 'f'
            value = float(value)
        elif type == 'bool':
            value = str(bool(value)).lower()
        elif type.startswith('uint'):
            suffix = 'U'
            value = int(round(value))
        else:
            value = int(round(value))

        return str(value) + suffix

    if parameter.default is None:
        meta_default = 0
    else:
        meta_default = parameter.default
    meta_default = create_literal(
        value=meta_default,
        type=parameter.internal_type,
    )

    if parameter.minimum is None:
        meta_minimum = types[parameter.internal_type].minimum_code
    else:
        meta_minimum = parameter.minimum
        meta_minimum = create_literal(
            value=meta_minimum,
            type=parameter.internal_type,
        )

    if parameter.maximum is None:
        meta_maximum = types[parameter.internal_type].maximum_code
    else:
        meta_maximum = parameter.maximum
        meta_maximum = create_literal(
            value=meta_maximum,
            type=parameter.internal_type,
        )

    meta_initializer_values = [
        f'[Meta_UserDefault - 1] = {meta_default},',
        f'[Meta_FactoryDefault - 1] = {meta_default},',
        f'[Meta_Min - 1] = {meta_minimum},',
        f'[Meta_Max - 1] = {meta_maximum}',
    ]
    return meta_initializer_values


def get_access_level_string(parameter, parameter_uuid_finder):
    access_level_uuid = parameter.access_level_uuid

    try:
        access_level = parameter_uuid_finder(access_level_uuid)
    except epyqlib.attrsmodel.NotFoundError as e:
        raise InvalidAccessLevelError.build(
            value=access_level_uuid,
            parameter=parameter,
        ) from e

    access_level = f'CAN_Enum_AccessLevel_{access_level.name}'

    return access_level


def can_getter_setter_variable(can_signal, parameter, var_or_func_or_table):
    if can_signal is None:
        can_variable = 'NULL'
        can_getter = 'NULL'
        can_setter = 'NULL'

        return can_getter, can_setter, can_variable

    in_table = isinstance(
        can_signal.tree_parent.tree_parent,
        epcpm.canmodel.CanTable,
    )
    if in_table:
        can_variable = (
            f'&{can_signal.tree_parent.tree_parent.tree_parent.name}'
            f'.{can_signal.tree_parent.tree_parent.name}'
            f'{can_signal.tree_parent.name}'
            f'.{can_signal.name}'
        )
    else:
        can_variable = (
            f'&{can_signal.tree_parent.tree_parent.name}'
            f'.{can_signal.tree_parent.name}'
            f'.{can_signal.name}'
        )

    if can_signal.signed:
        can_type = ''
    else:
        can_type = 'u'

    can_type += 'int'

    if can_signal.bits <= 16:
        can_type += '16'
    elif can_signal.bits <= 32:
        can_type += '32'
    else:
        raise Exception('ack')

    can_type += '_t'

    getter_setter_list = [
        'InterfaceItem',
        var_or_func_or_table,
        parameter.internal_type,
        'can',
        can_type,
    ]

    can_getter = '_'.join(
        str(x)
        for x in getter_setter_list + ['getter']
    )
    can_setter = '_'.join(
        str(x)
        for x in getter_setter_list + ['setter']
    )

    return can_getter, can_setter, can_variable


# TODO: CAMPid 68945967541316743769675426795146379678431
def breakdown_nested_array(s):
    split = re.split(r'\[(.*?)\].', s)

    array_layers = list(toolz.partition(2, split))
    remainder, = split[2 * len(array_layers):]

    return array_layers, remainder


# TODO: CAMPid 0974567213671436714671907842679364
@attr.s
class NestedArrays:
    array_layers = attr.ib()
    remainder = attr.ib()

    @classmethod
    def build(cls, s):
        array_layers, remainder = breakdown_nested_array(s)

        return cls(
            array_layers=array_layers,
            remainder=remainder,
        )

    def index(self, indexes):
        try:
            return '.'.join(
                '{layer}[{index}]'.format(
                    layer=layer,
                    index=index_format.format(**indexes),
                )
                for (layer, index_format), index in zip(self.array_layers, indexes)
            )
        except KeyError as e:
            raise

    def sizeof(self, layers):
        indexed = self.index(indexes={layer: 0 for layer in layers})

        return f'sizeof({indexed})'

    # def sizeof(self, layers, remainder=False):
    #     indexed = self.index(indexes={layer: 0 for layer in layers})

    #     if remainder:
    #         if len(layers) != len(self.array_layers):
    #             raise Exception('Remainder requested without specifying all layers')

    #         indexed += f'.{self.remainder}'

    #     return f'sizeof({indexed})'


@attr.s
class TableBaseStructures:
    array_nests = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()
    include_uuid_in_item = attr.ib()
    common_structure_names = attr.ib(factory=dict)
    c_code = attr.ib(factory=list)
    h_code = attr.ib(factory=list)

    def ensure_common_structure(
            self,
            internal_type,
            parameter_uuid,
            remainder,
            common_initializers,
            meta_initializer,
            setter,
    ):
        name = self.common_structure_names.get(parameter_uuid)

        if name is None:
            if len(self.h_code) > 0:
                self.h_code.append('')
            if len(self.c_code) > 0:
                self.c_code.append('')

            formatted_uuid = str(parameter_uuid).replace('-', '_')
            name = (
                f'InterfaceItem_table_common_{internal_type}_{formatted_uuid}'
            )

            nested_array = self.array_nests['x']

            layers = []
            for layer in nested_array.array_layers:
                layer_format_name, = [
                    list(field)[0][1]
                    for field in [string.Formatter().parse(layer[1])]
                ]
                layers.append(layer_format_name)

            variable_base = nested_array.index(
                indexes={
                    layer: 0
                    for layer in layers
                },
            )

            sizes = [
                self.array_nests['x'].sizeof(layers[:i + 1])
                for i in range(len(layers))
            ]

            if len(sizes) == 3:
                zone_size, curve_size, point_size = sizes
            else:
                zone_size = 0
                curve_size, point_size = sizes

            self.common_structure_names[parameter_uuid] = name
            self.h_code.append(
                f'extern InterfaceItem_table_common_{internal_type} {name};',
            )
            self.c_code.extend([
                f'#pragma DATA_SECTION({name}, "Interface")',
                f'// {parameter_uuid}',
                f'InterfaceItem_table_common_{internal_type} const {name} = {{',
                [
                    f'.common = {{',
                    common_initializers,
                    f'}},',
                    f'.variable_base = &{variable_base}.{remainder},',
                    f'.setter = {setter},',
                    f'.zone_size = {zone_size},',
                    f'.curve_size = {curve_size},',
                    f'.point_size = {point_size},',
                    f'.meta_values = {{',
                    meta_initializer,
                    f'}},',
                ],
                f'}};',
            ])

        return name

    def create_item(self, table_element, layers, sunspec_point):
        # TODO: CAMPid 9655426754319431461354643167
        array_element = table_element.original

        if isinstance(array_element, epyqlib.pm.parametermodel.Parameter):
            parameter = array_element
        else:
            parameter = array_element.tree_parent.children[0]

        if parameter.internal_variable is None:
            return [[], []]

        curve_type = get_curve_type(''.join(layers[:2]))

        curve_index = int(layers[-2])
        point_index = int(table_element.name.lstrip('_').lstrip('0')) - 1

        access_level = get_access_level_string(
            parameter=parameter,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        can_signal = self.parameter_uuid_to_can_node.get(table_element.uuid)

        can_getter, can_setter, can_variable = can_getter_setter_variable(
            can_signal,
            parameter,
            var_or_func_or_table='table',
        )

        # TODO: CAMPid 954679654745154274579654265294624765247569765479
        sunspec_getter = 'NULL'
        sunspec_setter = 'NULL'
        sunspec_variable = 'NULL'
        scale_factor_variable = 'NULL'
        scale_factor_updater = 'NULL'

        if sunspec_point is not None:
            sunspec_type = sunspec_types[
                self.parameter_uuid_finder(sunspec_point.type_uuid).name
            ]

            # TODO: CAMPid 9675436715674367943196954756419543975314
            getter_setter_list = [
                'InterfaceItem',
                'table',
                parameter.internal_type,
                'sunspec',
                sunspec_type,
            ]

            node_in_model = get_sunspec_point_from_table_element(
                sunspec_point=sunspec_point,
                table_element=table_element,
            )

            if node_in_model is not None:
                model_id = node_in_model.tree_parent.tree_parent.id
                sunspec_model_variable = f'sunspecInterface.model{model_id}'
                abbreviation = table_element.abbreviation
                sunspec_variable = (
                    f'&{sunspec_model_variable}'
                    f'.Curve_{curve_index + 1:>02}_{abbreviation}'
                )

                sunspec_getter = '_'.join(
                    str(x)
                    for x in getter_setter_list + ['getter']
                )
                sunspec_setter = '_'.join(
                    str(x)
                    for x in getter_setter_list + ['setter']
                )

            # TODO: CAMPid 67549654267913467967436
            sunspec_scale_factor = None
            factor_uuid = None
            if node_in_model is not None:
                if node_in_model.factor_uuid is not None:
                    factor_uuid = node_in_model.factor_uuid

            if factor_uuid is not None:
                root = node_in_model.find_root()
                factor_point = root.model.node_from_uuid(
                    node_in_model.factor_uuid,
                )
                sunspec_scale_factor_node = self.parameter_uuid_finder(
                    factor_point.parameter_uuid,
                )
                sunspec_scale_factor = sunspec_scale_factor_node.abbreviation

            if sunspec_scale_factor is not None:
                scale_factor_variable = (
                    f'&{sunspec_model_variable}.{sunspec_scale_factor}'
                )
                scale_factor_updater_name = (
                    f'getSUNSPEC_MODEL{model_id}_{sunspec_scale_factor}'
                )
                scale_factor_updater = f'&{scale_factor_updater_name}'

        common_initializers = create_common_initializers(
            access_level=access_level,
            can_getter=can_getter,
            can_setter=can_setter,
            # not to be used so really hardcode NULL
            can_variable='NULL',
            hand_coded_sunspec_getter_function='NULL',
            hand_coded_sunspec_setter_function='NULL',
            internal_scale=parameter.internal_scale_factor,
            scale_factor_updater=scale_factor_updater,
            scale_factor_variable=scale_factor_variable,
            sunspec_getter=sunspec_getter,
            sunspec_setter=sunspec_setter,
            # not to be used so really hardcode NULL
            sunspec_variable='NULL',
            can_scale_factor=can_signal.factor,
            uuid_=table_element.uuid,
            include_uuid_in_item=self.include_uuid_in_item,
        )

        meta_initializer = create_meta_initializer_values(parameter)

        remainder = NestedArrays.build(parameter.internal_variable).remainder

        common_structure_name = self.ensure_common_structure(
            internal_type=parameter.internal_type,
            parameter_uuid=parameter.uuid,
            remainder=remainder,
            common_initializers=common_initializers,
            meta_initializer=meta_initializer,
            setter=parameter.setter_function,
        )

        interface_item_type = (
            f'InterfaceItem_table_{parameter.internal_type}'
        )

        item_uuid_string = str(table_element.uuid).replace('-', '_')
        item_name = f'interfaceItem_{item_uuid_string}'

        maybe_uuid = []
        if self.include_uuid_in_item:
            maybe_uuid = [f'.uuid = {uuid_initializer(table_element.uuid)},']

        c = [
            f'#pragma DATA_SECTION({item_name}, "Interface")',
            f'// {table_element.uuid}',
            f'{interface_item_type} const {item_name} = {{',
            [
                f'.table_common = &{common_structure_name},',
                f'.can_variable = {can_variable},',
                f'.sunspec_variable = {sunspec_variable},',
                f'.zone = {curve_type if curve_type is not None else "0"},',
                f'.curve = {curve_index},',
                f'.point = {point_index},',
                *maybe_uuid,
            ],
            '};',
            '',
        ]

        return [
            c,
            [f'extern {interface_item_type} const {item_name};'],
        ]


# TODO: CAMPid 3078980986754174316996743174316967431
def get_sunspec_point_from_table_element(sunspec_point, table_element):
    value = table_element.original

    if isinstance(value, epyqlib.pm.parametermodel.ArrayParameterElement):
        value = value.original

    value = value.uuid

    nodes_in_model = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=value,
            attribute_name='parameter_uuid',
            raise_=False,
        )
        if isinstance(
            node,
            epcpm.sunspecmodel.TableRepeatingBlockReferenceDataPointReference,
        )
    ]

    for node in nodes_in_model:
        for child in node.tree_parent.original.children:
            if child.parameter_uuid == sunspec_point.parameter_uuid:
                node_in_model = node
                break
        else:
            continue

        break
    else:
        node_in_model = None
    return node_in_model


# TODO: CAMPid 3078980986754174316996743174316967431
def get_sunspec_model_from_table_group_element(sunspec_point, table_element):
    nodes_in_model = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=table_element.uuid,
            attribute_name='parameter_uuid',
            raise_=False,
        )
        if isinstance(node, epcpm.sunspecmodel.DataPoint)
    ]

    for node in nodes_in_model:
        for child in node.tree_parent.children:
            if child.parameter_uuid == table_element.uuid:
                node_in_model = node
                break
        else:
            continue

        break
    else:
        return None

    model_repeating_block, = [
        node
        for node in sunspec_point.find_root().nodes_by_attribute(
            attribute_value=node_in_model.tree_parent,
            attribute_name='original',
            raise_=False,
        )
        if isinstance(node, epcpm.sunspecmodel.TableRepeatingBlockReference)
    ]

    return model_repeating_block.tree_parent


@builders(epyqlib.pm.parametermodel.Table)
@attr.s
class Table:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        group, = (
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.TableGroupElement)
        )

        arrays = [
            child
            for child in self.wrapped.children
            if isinstance(child, epyqlib.pm.parametermodel.Array)
        ]

        # TODO: CAMPid 0795436754762451671643967431
        # TODO: get this from the ...  wherever we have it
        axes = ['x', 'y', 'z']

        array_nests = {
            name: NestedArrays.build(s=array.children[0].internal_variable)
            for name, array in zip(axes, arrays)
        }

        table_base_structures = TableBaseStructures(
            array_nests=array_nests,
            parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
            parameter_uuid_to_sunspec_node=(
                self.parameter_uuid_to_sunspec_node
            ),
            parameter_uuid_finder=self.parameter_uuid_finder,
            include_uuid_in_item=self.include_uuid_in_item,
        )

        item_code = builders.wrap(
            wrapped=group,
            can_root=self.can_root,
            sunspec_root=self.sunspec_root,
            table_base_structures=table_base_structures,
            parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
            parameter_uuid_to_sunspec_node=(
                self.parameter_uuid_to_sunspec_node
            ),
            parameter_uuid_finder=self.parameter_uuid_finder,
            include_uuid_in_item=self.include_uuid_in_item,
        ).gen()

        return [
            [
                *table_base_structures.c_code,
                '',
                *item_code[0],
            ],
            [
                *table_base_structures.h_code,
                '',
                *item_code[1],
            ],
        ]


@builders(epyqlib.pm.parametermodel.TableGroupElement)
@attr.s
class TableGroupElement:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    table_base_structures = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()
    layers = attr.ib(default=[])

    def gen(self):
        c = []
        h = []

        table_tree_root = not isinstance(
            self.wrapped.tree_parent,
            epyqlib.pm.parametermodel.TableGroupElement,
        )

        layers = list(self.layers)
        if not table_tree_root:
            layers.append(self.wrapped.name)

        for child in self.wrapped.children:
            result = builders.wrap(
                wrapped=child,
                can_root=self.can_root,
                sunspec_root=self.sunspec_root,
                table_base_structures=self.table_base_structures,
                parameter_uuid_to_can_node=self.parameter_uuid_to_can_node,
                parameter_uuid_to_sunspec_node=(
                    self.parameter_uuid_to_sunspec_node
                ),
                parameter_uuid_finder=self.parameter_uuid_finder,
                layers=layers,
                include_uuid_in_item=self.include_uuid_in_item,
            ).gen()

            c_built, h_built = result
            c.extend(c_built)
            h.extend(h_built)

        return c, h


# TODO: CAMPid 079549750417808543178043180
def get_curve_type(combination_string):
    # TODO: backmatching
    return {
        'LowRideThrough': 'IEEE1547_CURVE_TYPE_LRT',
        'HighRideThrough': 'IEEE1547_CURVE_TYPE_HRT',
        'LowTrip': 'IEEE1547_CURVE_TYPE_LTRIP',
        'HighTrip': 'IEEE1547_CURVE_TYPE_HTRIP',
    }.get(combination_string)


@builders(epyqlib.pm.parametermodel.TableArrayElement)
@attr.s
class TableArrayElement:
    wrapped = attr.ib()
    can_root = attr.ib()
    sunspec_root = attr.ib()
    table_base_structures = attr.ib()
    layers = attr.ib()
    include_uuid_in_item = attr.ib()
    parameter_uuid_to_can_node = attr.ib()
    parameter_uuid_to_sunspec_node = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        table_element = self.wrapped

        # TODO: CAMPid 9655426754319431461354643167
        array_element = table_element.original

        if isinstance(array_element, epyqlib.pm.parametermodel.Parameter):
            parameter = array_element
        else:
            parameter = array_element.tree_parent.children[0]

        is_group = isinstance(
            parameter.tree_parent,
            epyqlib.pm.parametermodel.Group,
        )

        if is_group:
            return self.handle_group()

        return self.handle_array()

    def handle_array(self):
        table_element = self.wrapped
        zone_node = table_element.tree_parent.tree_parent.tree_parent
        curve_node = zone_node.children[0]
        parameter = curve_node.descendent(
            self.wrapped.tree_parent.name,
            self.wrapped.name,
        )

        sunspec_point = self.parameter_uuid_to_sunspec_node.get(parameter.uuid)

        return self.table_base_structures.create_item(
            table_element=self.wrapped,
            layers=self.layers,
            sunspec_point=sunspec_point,
        )

    def handle_group(self):
        # raise Exception('...')

        table_element = self.wrapped
        zone_node = table_element.tree_parent.tree_parent.tree_parent
        curve_node = zone_node.children[0]
        axis_node = curve_node.descendent(*self.layers[1:])
        curve_0_table_element = axis_node.descendent(table_element.name)

        curve_index = int(self.layers[-2])

        parameter = table_element.original

        if parameter.internal_variable is None:
            return [[], []]

        can_signal = self.parameter_uuid_to_can_node.get(table_element.uuid)

        access_level = get_access_level_string(
            parameter=table_element,
            parameter_uuid_finder=self.parameter_uuid_finder,
        )

        # axis = table_element.tree_parent.axis

        # if parameter.setter_function is None:
        #     setter_function = 'NULL'
        # else:
        #     setter_function = parameter.setter_function.format(
        #         upper_axis=axis.upper(),
        #     )

        if parameter.setter_function is None:
            setter_function = 'NULL'
        else:
            setter_function = '&' + parameter.setter_function

        curve_type = get_curve_type(''.join(self.layers[:2]))

        internal_variable = parameter.internal_variable.format(
            curve_type=curve_type,
            curve_index=curve_index,
        )

        meta_initializer = create_meta_initializer_values(parameter)

        variable_or_getter_setter = [
            f'.variable = &{internal_variable},',
            f'.setter = {setter_function},',
            f'.meta_values = {{',
            meta_initializer,
            f'}},',
        ]

        # var_or_func = 'variable'

        can_getter, can_setter, can_variable = can_getter_setter_variable(
            can_signal,
            parameter,
            var_or_func_or_table='variable',
        )

        interface_item_type = (
            f'InterfaceItem_variable_{parameter.internal_type}'
        )

        # signal = self.parameter_uuid_to_can_node.get(self.wrapped.uuid)
        #
        # if signal is None:
        #     return None
        #
        # message = signal.tree_parent
        #
        # can_table = message.tree_parent

        # can_getter_setter_base = '_'.join(
        #     'InterfaceItem',
        #     'table',
        #     parameter.internal_type,
        #     'can',
        #     signal.can_interface_type,
        # )

        # can_getter, can_setter, can_variable = can_getter_setter_variable(
        #     can_signal,
        #     parameter,
        #     var_or_func_or_table=var_or_func,
        # )

        # TODO: CAMPid 954679654745154274579654265294624765247569765479
        sunspec_getter = 'NULL'
        sunspec_setter = 'NULL'
        sunspec_variable = 'NULL'

        sunspec_point = self.parameter_uuid_to_sunspec_node.get(
            table_element.original.uuid,
        )

        if sunspec_point is not None:
            sunspec_type = sunspec_types[
                self.parameter_uuid_finder(sunspec_point.type_uuid).name
            ]

            # TODO: CAMPid 9675436715674367943196954756419543975314
            getter_setter_list = [
                'InterfaceItem',
                'variable',
                parameter.internal_type,
                sunspec_type,
            ]
            print(end='')
            model = get_sunspec_model_from_table_group_element(
                sunspec_point=sunspec_point,
                table_element=curve_0_table_element,
            )

            if model is not None:
                model_id = model.id
                sunspec_model_variable = f'sunspecInterface.model{model_id}'
                abbreviation = parameter.abbreviation
                sunspec_variable = (
                    f'&{sunspec_model_variable}'
                    f'.Curve_{curve_index + 1:>02}_{abbreviation}'
                )

                sunspec_getter = '_'.join(
                    str(x)
                    for x in getter_setter_list + ['getter']
                )
                sunspec_setter = '_'.join(
                    str(x)
                    for x in getter_setter_list + ['setter']
                )

        result = create_item(
            item_uuid=table_element.uuid,
            access_level=access_level,
            can_getter=can_getter,
            can_setter=can_setter,
            can_variable=can_variable,
            hand_coded_sunspec_getter_function='NULL',
            hand_coded_sunspec_setter_function='NULL',
            interface_item_type=interface_item_type,
            internal_scale=parameter.internal_scale_factor,
            meta_initializer_values=create_meta_initializer_values(parameter),
            parameter=parameter,
            scale_factor_updater='NULL',
            scale_factor_variable='NULL',
            sunspec_getter=sunspec_getter,
            sunspec_setter=sunspec_setter,
            sunspec_variable=sunspec_variable,
            variable_or_getter_setter=variable_or_getter_setter,
            can_scale_factor=getattr(can_signal, 'factor', None),
            include_uuid_in_item=self.include_uuid_in_item,
        )

        return result


def create_item(
        item_uuid,
        include_uuid_in_item,
        access_level,
        can_getter,
        can_setter,
        can_variable,
        hand_coded_sunspec_getter_function,
        hand_coded_sunspec_setter_function,
        interface_item_type,
        internal_scale,
        meta_initializer_values,
        parameter,
        scale_factor_updater,
        scale_factor_variable,
        sunspec_getter,
        sunspec_setter,
        sunspec_variable,
        variable_or_getter_setter,
        can_scale_factor,
):
    item_uuid_string = str(item_uuid).replace('-', '_')
    item_name = f'interfaceItem_{item_uuid_string}'

    if meta_initializer_values is None:
        meta_initializer = []
    else:
        meta_initializer = [
            '.meta_values = {',
            meta_initializer_values,
            '}',
        ]

    common_initializers = create_common_initializers(
        access_level=access_level,
        can_getter=can_getter,
        can_setter=can_setter,
        can_variable=can_variable,
        hand_coded_sunspec_getter_function=hand_coded_sunspec_getter_function,
        hand_coded_sunspec_setter_function=hand_coded_sunspec_setter_function,
        internal_scale=internal_scale,
        scale_factor_updater=scale_factor_updater,
        scale_factor_variable=scale_factor_variable,
        sunspec_getter=sunspec_getter,
        sunspec_setter=sunspec_setter,
        sunspec_variable=sunspec_variable,
        can_scale_factor=can_scale_factor,
        uuid_=item_uuid,
        include_uuid_in_item=include_uuid_in_item,
    )

    item = [
        f'#pragma DATA_SECTION({item_name}, "Interface")',
        f'// {item_uuid}',
        f'{interface_item_type} const {item_name} = {{',
        [
            '.common = {',
            common_initializers,
            '},',
            *variable_or_getter_setter,
            *meta_initializer,
        ],
        '};',
        '',
    ]

    return [
        item,
        [f'extern {interface_item_type} const {item_name};'],
    ]


def uuid_initializer(uuid_):
    return '{{{}}}'.format(
        ', '.join(
            '0x{:02x}{:02x}'.format(high, low)
            for low, high in toolz.partition_all(2, uuid_.bytes)
        ),
    )


def create_common_initializers(
        access_level,
        can_getter,
        can_setter,
        can_variable,
        hand_coded_sunspec_getter_function,
        hand_coded_sunspec_setter_function,
        internal_scale,
        scale_factor_updater,
        scale_factor_variable,
        sunspec_getter,
        sunspec_setter,
        sunspec_variable,
        can_scale_factor,
        uuid_,
        include_uuid_in_item,
):
    if can_scale_factor is None:
        # TODO: don't default here?
        can_scale_factor = 1

    maybe_uuid = []
    if include_uuid_in_item:
        maybe_uuid = [f'.uuid = {uuid_initializer(uuid_)},']

    common_initializers = [
        f'.sunspecScaleFactor = {scale_factor_variable},',
        f'.canScaleFactor = {float(can_scale_factor)}f,',
        f'.scaleFactorUpdater = {scale_factor_updater},',
        f'.internalScaleFactor = {internal_scale},',
        f'.sunspec = {{',
        [
            f'.variable = {sunspec_variable},',
            f'.getter = {sunspec_getter},',
            f'.setter = {sunspec_setter},',
            f'.handGetter = {hand_coded_sunspec_getter_function},',
            f'.handSetter = {hand_coded_sunspec_setter_function},',
        ],
        f'}},',
        f'.can = {{',
        [
            f'.variable = {can_variable},',
            f'.getter = {can_getter},',
            f'.setter = {can_setter},',
            f'.handGetter = NULL,',
            f'.handSetter = NULL,',
        ],
        f'}},',
        f'.access_level = {access_level},',
        *maybe_uuid,
    ]
    return common_initializers



import attr
import epyqlib.utils.general

import epcpm.canmodel


builders = epyqlib.utils.general.TypeMap()


def export(path, can_model):
    builder = builders.wrap(
        wrapped=can_model.root,
        parameter_uuid_finder=can_model.node_from_uuid,
        path=path,
    )

    path.parent.mkdir(parents=True, exist_ok=True)
    builder.gen()


@builders(epcpm.canmodel.Root)
@attr.s
class Root:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()
    path = attr.ib(default=None)

    def gen(self):
        lines = [
            '#include <stdbool.h>',
            '',
            '#include "canInterfaceGen.h"',
            '#include "gridMonitor.h"',
            '#include "utils.h"',
            '#include "IQmathLib.h"',
            '',
            '',
        ]

        table_results = []

        parameter_query = self.wrapped.child_by_name('ParameterQuery')

        for child in parameter_query.children:
            if not isinstance(child, epcpm.canmodel.CanTable):
                continue

            builder = builders.wrap(
                wrapped=child,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )

            table_results.append(builder.gen())

            lines.extend(table_results[-1].table_lines)
            lines.append('')

        active_curves = parameter_query.child_by_name('ActiveCurves')

        for get_or_set in ('get', 'set'):
            active_lines = []
            for table_result in table_results:
                active_lines.extend(
                    table_result.active_curves_lines[get_or_set],
                )

            lines.extend([
                f'void {get_or_set}{active_curves.name}(void)',
                '{',
                active_lines,
                '}',
                '',
            ])

        content = epyqlib.utils.general.format_nested_lists(lines)

        with self.path.open('w', newline='\n') as f:
            f.write(content)
        return


@attr.s
class TableResults:
    table_lines = attr.ib()
    active_curves_lines = attr.ib()


@builders(epcpm.canmodel.CanTable)
@attr.s
class CanTable:
    wrapped = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self):
        parameter_table = self.parameter_uuid_finder(
            self.wrapped.table_uuid,
        )

        return TableResults(
            table_lines=self.table_lines(parameter_table=parameter_table),
            active_curves_lines={
                'get': self.active_curves(
                    format_string=parameter_table.active_curve_getter,
                    parameter_table=parameter_table,
                ),
                'set': self.active_curves(
                    format_string=parameter_table.active_curve_setter,
                    parameter_table=parameter_table,
                ),
            }
        )

    def table_lines(self, parameter_table):
        lines = []

        can_getter = parameter_table.can_getter
        can_setter = parameter_table.can_setter

        for child in self.wrapped.children:
            if not isinstance(child, epcpm.canmodel.Multiplexer):
                continue

            builder = builders.wrap(
                wrapped=child,
                table_name=self.wrapped.name,
                can_getter=can_getter,
                can_setter=can_setter,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )

            for get_or_set in ('get', 'set'):
                new_lines = builder.gen(get_or_set=get_or_set)
                if len(new_lines) > 0:
                    lines.extend(new_lines)
                    lines.append('')

        return lines

    def active_curves(self, format_string, parameter_table):
        lines = []

        parameter_query = self.wrapped.tree_parent
        active_curves = parameter_query.child_by_name('ActiveCurves')

        for combination in parameter_table.curve_group_combinations:
            curve_type = get_curve_type(''.join(x.name for x in combination))

            interface_signal = active_curves.child_by_name(
                (
                        self.wrapped.name
                        + ''.join(enumerator.name for enumerator in combination)
                ),
            )

            full_path = '.'.join((
                interface_signal.tree_parent.tree_parent.name,
                interface_signal.tree_parent.name,
                interface_signal.name,
            ))

            lines.append(format_string.format(
                interface_signal=full_path,
                curve_type=curve_type,
            ))

        return lines


# TODO: CAMPid 079549750417808543178043180
def get_curve_type(combination_string):
    # TODO: backmatching
    return {
        'LowRideThrough': 'IEEE1547_CURVE_TYPE_LRT',
        'HighRideThrough': 'IEEE1547_CURVE_TYPE_HRT',
        'LowTrip': 'IEEE1547_CURVE_TYPE_LTRIP',
        'HighTrip': 'IEEE1547_CURVE_TYPE_HTRIP',
    }.get(combination_string)


@builders(epcpm.canmodel.Multiplexer)
@attr.s
class Multiplexer:
    wrapped = attr.ib()
    table_name = attr.ib()
    can_getter = attr.ib()
    can_setter = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self, get_or_set):
        children_lines = []

        curve_group_string, _, _ = self.wrapped.name.partition('_')
        curve_type = get_curve_type(curve_group_string)

        for child in self.wrapped.children:
            builder = builders.wrap(
                wrapped=child,
                table_name=self.table_name,
                can_getter=self.can_getter,
                can_setter=self.can_setter,
                multiplexer_name=self.wrapped.name,
                curve_type=curve_type,
                parameter_uuid_finder=self.parameter_uuid_finder,
            )

            children_lines.extend(builder.gen(
                get_or_set=get_or_set,
            ))

        if len(children_lines) > 0 and all(line is None for line in children_lines):
            return []

        children_lines = [
            line
            for line in children_lines
            if line is not None
        ]

        lines = [
            'void {get_or_set}{table}{mux}(void)'.format(
                get_or_set=get_or_set,
                table=self.table_name,
                mux=self.wrapped.name,
            ),
            '{',
            children_lines,
            '}'
        ]

        return lines

    def point_count(self):
        return len(self.wrapped.children)


@builders(epcpm.canmodel.Signal)
@attr.s
class Signal:
    wrapped = attr.ib()
    table_name = attr.ib()
    multiplexer_name = attr.ib()
    can_getter = attr.ib()
    can_setter = attr.ib()
    curve_type = attr.ib()
    parameter_uuid_finder = attr.ib()

    def gen(self, get_or_set):
        interface_signal = 'ParameterQuery.{table}{mux}.{signal}'.format(
            table=self.table_name,
            mux=self.multiplexer_name,
            signal=self.wrapped.name,
        )

        table_array_element = self.parameter_uuid_finder(
            self.wrapped.parameter_uuid
        )

        original = table_array_element
        while not isinstance(original, epyqlib.pm.parametermodel.Parameter):
            original = original.original

        if original.uses_interface_item():
            return [None]

        if isinstance(original.tree_parent, epyqlib.pm.parametermodel.Group):
            can_getter = original.can_getter
            can_setter = original.can_setter
        elif isinstance(original.tree_parent, epyqlib.pm.parametermodel.Array):
            can_getter = self.can_getter
            can_setter = self.can_setter

        if can_getter is None:
            can_getter = ''

        if can_setter is None:
            can_setter = ''

        axis = table_array_element.tree_parent.axis
        if axis is None:
            axis = '<no axis>'

        curve_index = table_array_element.tree_parent.tree_parent.curve_index

        lines = [
            {
                'get': can_getter,
                'set': can_setter,
            }[get_or_set].format(
                curve_type=self.curve_type,
                interface_signal=interface_signal,
                point_index=table_array_element.index,
                axis=axis,
                upper_axis=axis.upper(),
                curve_index=curve_index,
            )
        ]

        lines = [
            line
            for line in lines
            if len(line.strip()) > 0
        ]

        return lines



import pathlib

import alqtendpy.compileui
import setuptools
import versioneer


alqtendpy.compileui.compile_ui(
    directory_paths=[pathlib.Path(__file__).parent  / 'src' / 'epcpm'],
)


setuptools.setup(
    name='epcpm',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="EPC Power Corp.",
    classifiers=[
        ("License :: OSI Approved :: "
         "GNU General Public License v2 or later (GPLv2+)")
    ],
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    entry_points={
        'gui_scripts': [
            'epcpm = epcpm.__main__:_entry_point',
        ],
        'console_scripts': [
            'epcpmcli = epcpm.cli.main:main',
            'epcparameterstoc = epcpm.cli.parameterstoc:cli',
            'epcimportsym = epcpm.cli.importsym:cli',
            'epcexportsym = epcpm.cli.exportsym:cli',
            'epcexportdocx = epcpm.cli.exportdocx:cli',
            'epcconvertparameters = epcpm.cli.convertepp:cli',
        ],
    },
    install_requires=[
        'canmatrix>=0.9.1',
        'click',
        'epyqlib>=2019.12.10',
        'graham',
        'jinja2',
        'lxml',
        'openpyxl',
        'pycparser',
        'pysunspec',
        'pyqt5',
        'toolz',
        'xmldiff',
    ],
    extras_require={
        'test': [
            'codecov',
            'pytest',
            'pytest-cov',
            'pytest-qt',
            'tox',
        ],
        'build': [
            'pyinstaller',
        ]
    },
)



try:
    from epcpm.__build import (
        build_system,
        build_id,
        build_number,
        build_version,
        job_id,
        job_url,
    )
except ImportError:
    build_system = None
    build_id = None
    build_number = None
    build_version = None
    job_id = None
    job_url = None
